#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "global.h"
#include "options.h"
#include "macros.h"
#include "ali.h"
#include "ali3core.h"
#include "scores.h"
#include "util.h"
#include "sinfo.h"

/*********************************************************
returns the number of maximum argument of seven floats
**********************************************************/
static int maxarg7float (float a0, float a1, float a2, float a3, float a4, float a5, float a6)
{
  int i, maxarg;
  static float array[7];

  maxarg=0;
  array[0]=a0;
  array[1]=a1;
  array[2]=a2;
  array[3]=a3;
  array[4]=a4;
  array[5]=a5;
  array[6]=a6;

  for (i=1; i<7; i++) if (array[i]>array[maxarg]) maxarg=i;

  return (maxarg);
}

/*********************************************************
returns the maximum of six floats
**********************************************************/
static float max6float (float a0, float a1, float a2, float a3, float a4, float a5)
{
  int i;
  float max;
  static float array[6];

  array[0]=a0;
  array[1]=a1;
  array[2]=a2;
  array[3]=a3;
  array[4]=a4;
  array[5]=a5;
  max=array[0];

  for (i=1; i<6; i++) if (array[i]>max) max=array[i];

  return (max);
}

/*********************************************************
returns the maximum of seven floats
**********************************************************/
static float max7float (float a0, float a1, float a2, float a3, float a4, float a5, float a6)
{
  int i;
  float max;
  static float array[7];

  array[0]=a0;
  array[1]=a1;
  array[2]=a2;
  array[3]=a3;
  array[4]=a4;
  array[5]=a5;
  array[6]=a6;
  max=array[0];

  for (i=1; i<7; i++) if (array[i]>max) max=array[i];

  return (max);
}

/********************************************************
pairwise align a pair of the tree sequences given
to obtain faces of DP matrices needed for initialization
********************************************************/
static void DP3Dfaces (float ***DPscore, float ***Ixy, float ***Ixz, float ***Iyz, float ***Ix, float ***Iy, float ***Iz, char ***DPdir, 
		       float **Sxy, float **Sxz, float **Syz,
		       int s1, int l1, int s2, int l2, int s3, int l3, SCORING *scoring)
{
  int i, j, k;
  float sc[3];
  int maxarg;
  float f_i, f_j, f_k;

  /* get data */
  int m1=l1+1;
  int m2=l2+1;
  int m3=l3+1;

  float go=scoring->go;
  float ge=scoring->ge;
  BOOL tgw=scoring->termgapweight;

  /* init */
  DPscore[0][0][0]=0.0f;
  Ix     [0][0][0]=-FLARGE;
  Iy     [0][0][0]=-FLARGE;
  Iz     [0][0][0]=-FLARGE;
  DPdir  [0][0][0]=STOP;
  for (i=1; i<m1; i++)
    {
      DPscore[i][0][0]=(tgw ? -go-(float)(i-1)*ge : 0.0f);
      Ix     [i][0][0]=(tgw ? -go-(float)(i-1)*ge : 0.0f);
      Iy     [i][0][0]=-FLARGE;
      Iz     [i][0][0]=-FLARGE;
      DPdir  [i][0][0]=DIRX;
    }
  for (j=1; j<m2; j++)
    {
      DPscore[0][j][0]=(tgw ? -go-(float)(j-1)*ge : 0.0f);
      Ix     [0][j][0]=-FLARGE;
      Iy     [0][j][0]=(tgw ? -go-(float)(j-1)*ge : 0.0f);
      Iz     [0][j][0]=-FLARGE;
      DPdir  [0][j][0]=DIRY;
    }
  for (k=1; k<m3; k++)
    {
      DPscore[0][0][k]=(tgw ? -go-(float)(k-1)*ge : 0.0f);
      Ix     [0][0][k]=-FLARGE;
      Iy     [0][0][k]=-FLARGE;
      Iz     [0][0][k]=(tgw ? -go-(float)(k-1)*ge : 0.0f);
      DPdir  [0][0][k]=DIRZ;
    }

  /* fill xy */
  for (i=1; i<m1; i++)
    {
      for (j=1; j<m2; j++)
	{
	  /* set pre-factors */
	  f_i=((!tgw) && (i==l1) ? 0.0f : 1.0f);
	  f_j=((!tgw) && (j==l2) ? 0.0f : 1.0f);
	  
	  /* update matrices */
	  Ixy[i][j][0]=DPscore[i-1][j-1][0]+Sxy[s1+i-1][s2+j-1];
	  Ix [i][j][0]=MAX (DPscore[i-1][j][0]-f_j*go, Ix[i-1][j][0]-f_j*ge);
	  Iy [i][j][0]=MAX (DPscore[i][j-1][0]-f_i*go, Iy[i][j-1][0]-f_i*ge);
	  sc[0]=Ixy[i][j][0];
	  sc[1]=Ix [i][j][0];
	  sc[2]=Iy [i][j][0];
	  maxarg=MAXARG3 (sc[0], sc[1], sc[2]);
	  DPscore[i][j][0]=sc[maxarg];
	  if (maxarg==0) DPdir[i][j][0]=DIRXY;
	  else if (maxarg==1) DPdir[i][j][0]=DIRX;
	  else DPdir[i][j][0]=DIRY;
	}
    }

  /* fill xz */
  for (i=1; i<m1; i++)
    {
      for (k=1; k<m3; k++)
	{
	  /* set pre-factors */
	  f_i=((!tgw) && (i==l1) ? 0.0f : 1.0f);
	  f_k=((!tgw) && (k==l3) ? 0.0f : 1.0f);

	  /* update matrices */
	  Ixz[i][0][k]=DPscore[i-1][0][k-1]+Sxz[s1+i-1][s3+k-1];
	  Ix [i][0][k]=MAX (DPscore[i-1][0][k]-f_k*go, Ix[i-1][0][k]-f_k*ge);
	  Iz [i][0][k]=MAX (DPscore[i][0][k-1]-f_i*go, Iz[i][0][k-1]-f_i*ge);
	  sc[0]=Ixz[i][0][k];
	  sc[1]=Ix [i][0][k];
	  sc[2]=Iz [i][0][k];
	  maxarg=MAXARG3 (sc[0], sc[1], sc[2]);
	  DPscore[i][0][k]=sc[maxarg];
	  if (maxarg==0) DPdir[i][0][k]=DIRXZ;
	  else if (maxarg==1) DPdir[i][0][k]=DIRX;
	  else DPdir[i][0][k]=DIRZ;
	}
    }

  /* fill yz */
  for (j=1; j<m2; j++)
    {
      for (k=1; k<m3; k++)
	{
	  /* set pre-factors */
	  f_j=((!tgw) && (j==l2) ? 0.0f : 1.0f);
	  f_k=((!tgw) && (k==l3) ? 0.0f : 1.0f);

	  /* update matrices */
	  Iyz[0][j][k]=DPscore[0][j-1][k-1]+Syz[s2+j-1][s3+k-1];
	  Iy [0][j][k]=MAX (DPscore[0][j-1][k]-f_k*go, Iy[0][j-1][k]-f_k*ge);
	  Iz [0][j][k]=MAX (DPscore[0][j][k-1]-f_j*go, Iz[0][j][k-1]-f_j*ge);
	  sc[0]=Iyz[0][j][k];
	  sc[1]=Iy [0][j][k];
	  sc[2]=Iz [0][j][k];
	  maxarg=MAXARG3 (sc[0], sc[1], sc[2]);
	  DPscore[0][j][k]=sc[maxarg];
	  if (maxarg==0) DPdir[0][j][k]=DIRYZ;
	  else if (maxarg==1) DPdir[0][j][k]=DIRY;
	  else DPdir[0][j][k]=DIRZ;
	}
    }
}

/********************************************************
initializes DP matrices
********************************************************/
static void DP3Dinit (float ***DPscore, float ***Ixy, float ***Ixz, float ***Iyz, float ***Ix, float ***Iy, float ***Iz, char ***DPdir,
		      float go, float ge, int l1, int l2, int l3, SCORING *scoring)
{
  int i, j, k;

  /* get data */
  int m1=l1+1;
  int m2=l2+1;
  int m3=l3+1;

  BOOL tgw=scoring->termgapweight;

  /* fill matrix front faces */
  /***************************/
  /* planes            */
  /* xy-plane: a-plane */
  /* gaps in z */
  for (i=1; i<m1; i++) for (j=1; j<m2; j++)
    {
      DPscore[i][j][0]+=(tgw ? -2.0f*go-(float)(i+j-2)*ge : 0.0f);
      Ixy    [i][j][0]+=(tgw ? -2.0f*go-(float)(i+j-2)*ge : 0.0f);
      Ix     [i][j][0]+=(tgw ? -2.0f*go-(float)(i+j-2)*ge : 0.0f);
      Iy     [i][j][0]+=(tgw ? -2.0f*go-(float)(i+j-2)*ge : 0.0f);
      Iz     [i][j][0]=-FLARGE;
      Ixz    [i][j][0]=-FLARGE;
      Iyz    [i][j][0]=-FLARGE;
    }
  /* xz-plane: b-plane */
  /* gaps in y */
  for (i=1; i<m1; i++) for (k=1; k<m3; k++)
    {
      DPscore[i][0][k]+=(tgw ? -2.0f*go-(float)(i+k-2)*ge : 0.0f);
      Ixz    [i][0][k]+=(tgw ? -2.0f*go-(float)(i+k-2)*ge : 0.0f);
      Ix     [i][0][k]+=(tgw ? -2.0f*go-(float)(i+k-2)*ge : 0.0f);
      Iz     [i][0][k]+=(tgw ? -2.0f*go-(float)(i+k-2)*ge : 0.0f);
      Iy     [i][0][k]=-FLARGE;
      Ixy    [i][0][k]=-FLARGE;
      Iyz    [i][0][k]=-FLARGE;
    }
  /* yz-plane: c-plane */
  /* gaps in x */
  for (j=1; j<m2; j++) for (k=1; k<m3; k++)
    {
      DPscore[0][j][k]+=(tgw ? -2.0f*go-(float)(j+k-2)*ge : 0.0f);
      Iyz    [0][j][k]+=(tgw ? -2.0f*go-(float)(j+k-2)*ge : 0.0f);
      Iy     [0][j][k]+=(tgw ? -2.0f*go-(float)(j+k-2)*ge : 0.0f);
      Iz     [0][j][k]+=(tgw ? -2.0f*go-(float)(j+k-2)*ge : 0.0f);
      Ix     [0][j][k]=-FLARGE;
      Ixy    [0][j][k]=-FLARGE;
      Ixz    [0][j][k]=-FLARGE;
    }
  /* axes */
  /* x-axis */
  /* gaps in y and z */
  for (i=1; i<m1; i++)
    {
      DPscore[i][0][0]+=(tgw ? -2.0f*go-2.0f*(float)(i-1)*ge : 0.0f);
      Ix     [i][0][0]+=(tgw ? -2.0f*go-2.0f*(float)(i-1)*ge : 0.0f);
      Iy     [i][0][0]=-FLARGE;
      Iz     [i][0][0]=-FLARGE;
      Ixy    [i][0][0]=-FLARGE;
      Ixz    [i][0][0]=-FLARGE;
      Iyz    [i][0][0]=-FLARGE;
      DPdir  [i][0][0]=DIRX;
    }
  /* y-axis */
  /* gaps in x and z */
  for (j=1; j<m2; j++)
    {
      DPscore[0][j][0]+=(tgw ? -2.0f*go-2.0f*(float)(j-1)*ge : 0.0f);
      Ix     [0][j][0]=-FLARGE;
      Iy     [0][j][0]+=(tgw ? -2.0f*go-2.0f*(float)(j-1)*ge : 0.0f);
      Iz     [0][j][0]=-FLARGE;
      Ixy    [0][j][0]=-FLARGE;
      Ixz    [0][j][0]=-FLARGE;
      Iyz    [0][j][0]=-FLARGE;
      DPdir  [0][j][0]=DIRY;
    }
  /* z-axis */
  /* gaps in x and y */
  for (k=1; k<m3; k++)
    {
      DPscore[0][0][k]+=(tgw ? -2.0f*go-2.0f*(float)(k-1)*ge : 0.0f);
      Ix     [0][0][k]=-FLARGE;
      Iy     [0][0][k]=-FLARGE;
      Iz     [0][0][k]+=(tgw ? -2.0f*go-2.0f*(float)(k-1)*ge : 0.0f);
      Ixy    [0][0][k]=-FLARGE;
      Ixz    [0][0][k]=-FLARGE;
      Iyz    [0][0][k]=-FLARGE;
      DPdir  [0][0][k]=DIRZ;
    }
  /* origin */
  DPscore[0][0][0]=0.0f;
  Ix     [0][0][0]=-FLARGE;
  Iy     [0][0][0]=-FLARGE;
  Iz     [0][0][0]=-FLARGE;
  Ixy    [0][0][0]=-FLARGE;
  Ixz    [0][0][0]=-FLARGE;
  Iyz    [0][0][0]=-FLARGE;
  DPdir  [0][0][0]=STOP;
}

/*******************************************************
align three alignments
********************************************************/
ALI *do3align (ALI *ali1, ALI *ali2, ALI *ali3, int s1, int l1, int s2, int l2, int s3, int l3, float **Sxy, float **Sxz, float **Syz, SCORING *scoring)
{
  int i, j, k;
  int len;
  int posi, posj, posk;
  float f_i, f_j, f_k;
  int maxscorearg;

  float ***DPscore, ***Ixy, ***Ixz, ***Iyz, ***Ix, ***Iy, ***Iz;
  char ***DPdir;
  char **newdes;
  char **newseqs;
  SINFO **newsinfo;
  ALI *ali;
  static const SINFO SINFOZERO={0.0f, 0.0f, 0.0f};

  /* get data */
  int m1=l1+1;
  int m2=l2+1;
  int m3=l3+1;
  int msum=m1+m2+m3;
  int n1=ali1->num;
  int n2=ali2->num;
  int n3=ali3->num;
  int num=n1+n2+n3;

  float go=scoring->go;
  float ge=scoring->ge;
  GAPTYPE gaptype=scoring->gaptype;
  BOOL tgw=scoring->termgapweight;
  
  /* allocate memory */
  MAKE3DFLOAT (DPscore,m1,m2,m3,i,j);
  MAKE3DFLOAT (Ixy,m1,m2,m3,i,j);
  MAKE3DFLOAT (Ixz,m1,m2,m3,i,j);
  MAKE3DFLOAT (Iyz,m1,m2,m3,i,j);
  MAKE3DFLOAT (Ix,m1,m2,m3,i,j);
  MAKE3DFLOAT (Iy,m1,m2,m3,i,j);
  MAKE3DFLOAT (Iz,m1,m2,m3,i,j);
  MAKE3DCHAR (DPdir,m1,m2,m3,i,j);

  /* initialize matrix */
  /*********************/
  /* align faces of DP matrices */
  DP3Dfaces (DPscore, Ixy, Ixz, Iyz, Ix, Iy, Iz, DPdir, Sxy, Sxz, Syz, s1, l1, s2, l2, s3, l3, scoring);

  /* initialize faces of DP matrices (align them first!) */
  DP3Dinit (DPscore, Ixy, Ixz, Iyz, Ix, Iy, Iz, DPdir, go, ge, l1, l2, l3, scoring);

  /* fill matrix */
  for (i=1; i<m1; i++)
    {
      for (j=1; j<m2; j++)
	{
	  for (k=1; k<m3; k++)
	    {
	      /* set pre-factors */
	      f_i=((!tgw) && (i==l1) ? 0.0f : 1.0f);
	      f_j=((!tgw) && (j==l2) ? 0.0f : 1.0f);
	      f_k=((!tgw) && (k==l3) ? 0.0f : 1.0f);

	      /* update DP matrices */
	      if (gaptype==GAP_SIMPLE) /* quasi-natural gap costs */
		{
		  /* DP */
		  /* no gaps */
		  DPscore[i][j][k]=DPscore[i-1][j-1][k-1] + Sxy[s1+i-1][s2+j-1] + Sxz[s1+i-1][s3+k-1] + Syz[s2+j-1][s3+k-1]; /* add the three pairwise costs */

#if 1             
		  /* standard */
		  /* Ixy, Ixz, Iyz */
		  /* gap in z */
		  Ixy[i][j][k]=MAX (DPscore[i-1][j-1][k] - f_k*2.0f*go, Ixy[i-1][j-1][k] - f_k*2.0f*ge) + Sxy[s1+i-1][s2+j-1];
		  /* gap in y */
		  Ixz[i][j][k]=MAX (DPscore[i-1][j][k-1] - f_j*2.0f*go, Ixz[i-1][j][k-1] - f_j*2.0f*ge) + Sxz[s1+i-1][s3+k-1];
		  /* gap in x */
		  Iyz[i][j][k]=MAX (DPscore[i][j-1][k-1] - f_i*2.0f*go, Iyz[i][j-1][k-1] - f_i*2.0f*ge) + Syz[s2+j-1][s3+k-1];
#else
		  /* Ixy, Ixz, Iyz */
		  /* gap in z */
		  Ixy[i][j][k]=MAX (DPscore[i-1][j-1][k] - f_k*go, Ixy[i-1][j-1][k] - f_k*ge) + Sxy[s1+i-1][s2+j-1];
		  /* gap in y */
		  Ixz[i][j][k]=MAX (DPscore[i-1][j][k-1] - f_j*go, Ixz[i-1][j][k-1] - f_j*ge) + Sxz[s1+i-1][s3+k-1];
		  /* gap in x */
		  Iyz[i][j][k]=MAX (DPscore[i][j-1][k-1] - f_i*go, Iyz[i][j-1][k-1] - f_i*ge) + Syz[s2+j-1][s3+k-1];
#endif
		  /* Ix, Iy, Iz */
		  /* gap in y and z */
		  Ix[i][j][k]=MAX (DPscore[i-1][j][k] - f_j*go-f_k*go, Ix[i-1][j][k] - f_j*ge-f_k*ge);
		  /* gap in x and z */
		  Iy[i][j][k]=MAX (DPscore[i][j-1][k] - f_i*go-f_k*go, Iy[i][j-1][k] - f_i*ge-f_k*ge);
		  /* gap in x and y */
		  Iz[i][j][k]=MAX (DPscore[i][j][k-1] - f_i*go-f_j*go, Iz[i][j][k-1] - f_i*ge-f_j*ge);
		}
	      else /* natural gap costs */
		{
#if 0
		  /* DP */
		  /* no gaps */
		  DPscore[i][j][k]=DPscore[i-1][j-1][k-1] + Sxy[s1+i-1][s2+j-1] + Sxz[s1+i-1][s3+k-1] + Syz[s2+j-1][s3+k-1];
		  /* Ixy, Ixz, Iyz */
		  /* gap in z */
		  Ixy[i][j][k]=max6float (DPscore[i-1][j-1][k] - f_k*2.0f*go,  /* no gap -> gap z */
					  Ixy    [i-1][j-1][k] - f_k*2.0f*ge,  /* gap z  -> gap z */
					  Ixz    [i-1][j-1][k] - f_k*2.0f*go,  /* gap y  -> gap z */
					  Iyz    [i-1][j-1][k] - f_k*2.0f*go,  /* gap x  -> gap z */
					  Ix     [i-1][j-1][k] - f_k*2.0f*ge,  /* gap yz -> gap z */
					  Iy     [i-1][j-1][k] - f_k*2.0f*ge)  /* gap xz -> gap z */
		    + Sxy[s1+i-1][s2+j-1];

		  /* gap in y */
		  Ixz[i][j][k]=max6float (DPscore[i-1][j][k-1] - f_j*2.0f*go,  /* no gap -> gap y */
					  Ixy    [i-1][j][k-1] - f_j*2.0f*go,  /* gap z  -> gap y */
					  Ixz    [i-1][j][k-1] - f_j*2.0f*ge,  /* gap y  -> gap y */
					  Iyz    [i-1][j][k-1] - f_j*2.0f*go,  /* gap x  -> gap y */
					  Ix     [i-1][j][k-1] - f_j*2.0f*ge,  /* gap yz -> gap y */
					  Iz     [i-1][j][k-1] - f_j*2.0f*ge)  /* gap xy -> gap y */
		    + Sxz[s1+i-1][s3+k-1];

		  /* gap in x */
		  Iyz[i][j][k]=max6float (DPscore[i][j-1][k-1] - f_i*2.0f*go,  /* no gap -> gap x */
					  Ixy    [i][j-1][k-1] - f_i*2.0f*go,  /* gap z  -> gap x */
					  Ixz    [i][j-1][k-1] - f_i*2.0f*go,  /* gap y  -> gap x */
					  Iyz    [i][j-1][k-1] - f_i*2.0f*ge,  /* gap x  -> gap x */
					  Iy     [i][j-1][k-1] - f_i*2.0f*ge,  /* gap xz -> gap x */
					  Iz     [i][j-1][k-1] - f_i*2.0f*ge)  /* gap xy -> gap x */
		    + Syz[s2+j-1][s3+k-1];

		  /* Ix, Iy, Iz */
		  /* gaps in y and z */
		  Ix[i][j][k]=max6float (DPscore[i-1][j][k] - f_j*go-f_k*go,  /* no gap -> gap yz */
					 Ixy    [i-1][j][k] - f_j*go-f_k*ge,  /* gap z  -> gap yz */
					 Ixz    [i-1][j][k] - f_j*ge-f_k*go,  /* gap y  -> gap yz */
					 Ix     [i-1][j][k] - f_j*ge-f_k*ge,  /* gap yz -> gap yz */
					 Iy     [i-1][j][k] - f_j*go-f_k*ge,  /* gap xz -> gap yz */
					 Iz     [i-1][j][k] - f_j*ge-f_k*go); /* gap xy -> gap yz */

		  /* gaps in x and z */
		  Iy[i][j][k]=max6float (DPscore[i][j-1][k] - f_i*go-f_k*go,  /* no gap -> gap xz */
					 Ixy    [i][j-1][k] - f_i*go-f_k*ge,  /* gap z  -> gap xz */
					 Iyz    [i][j-1][k] - f_i*ge-f_k*go,  /* gap x  -> gap xz */
					 Ix     [i][j-1][k] - f_i*go-f_k*ge,  /* gap yz -> gap xz */
					 Iy     [i][j-1][k] - f_i*ge-f_k*ge,  /* gap xz -> gap xz */
					 Iz     [i][j-1][k] - f_i*ge-f_k*go); /* gap xy -> gap xz */

		  /* gaps in x and y */
		  Iz[i][j][k]=max6float (DPscore[i][j][k-1] - f_i*go-f_j*go,  /* no gap -> gap xy */
					 Ixz    [i][j][k-1] - f_i*go-f_j*ge,  /* gap y  -> gap xy */
					 Iyz    [i][j][k-1] - f_i*ge-f_j*go,  /* gap x  -> gap xy */
					 Ix     [i][j][k-1] - f_i*go-f_j*ge,  /* gap yz -> gap xy */
					 Iy     [i][j][k-1] - f_i*ge-f_j*go,  /* gap xz -> gap xy */
					 Iz     [i][j][k-1] - f_i*ge-f_j*ge); /* gap xy -> gap xy */
#else
		  /* DP */
		  /* no gaps */
		  DPscore[i][j][k]=DPscore[i-1][j-1][k-1] + Sxy[s1+i-1][s2+j-1] + Sxz[s1+i-1][s3+k-1] + Syz[s2+j-1][s3+k-1];
		  /* Ixy, Ixz, Iyz */
		  /* gap in z */
		  Ixy[i][j][k]=max7float (DPscore[i-1][j-1][k] - f_k*go,  /* no gap -> gap z */
					  Ixy    [i-1][j-1][k] - f_k*ge,  /* gap z  -> gap z */
					  Ixz    [i-1][j-1][k] - f_k*go,  /* gap y  -> gap z */
					  Iyz    [i-1][j-1][k] - f_k*go,  /* gap x  -> gap z */
					  Ix     [i-1][j-1][k] - f_k*ge,  /* gap yz -> gap z */
					  Iy     [i-1][j-1][k] - f_k*ge,  /* gap xz -> gap z */
					  Iz     [i-1][j-1][k] - f_k*go)  /* gap xy -> gap z */
		    + Sxy[s1+i-1][s2+j-1];

		  /* gap in y */
		  Ixz[i][j][k]=max7float (DPscore[i-1][j][k-1] - f_j*go,  /* no gap -> gap y */
					  Ixy    [i-1][j][k-1] - f_j*go,  /* gap z  -> gap y */
					  Ixz    [i-1][j][k-1] - f_j*ge,  /* gap y  -> gap y */
					  Iyz    [i-1][j][k-1] - f_j*go,  /* gap x  -> gap y */
					  Ix     [i-1][j][k-1] - f_j*ge,  /* gap yz -> gap y */
					  Iy     [i-1][j][k-1] - f_j*go,  /* gap xz -> gap y */
					  Iz     [i-1][j][k-1] - f_j*ge)  /* gap xy -> gap y */
		    + Sxz[s1+i-1][s3+k-1];

		  /* gap in x */
		  Iyz[i][j][k]=max7float (DPscore[i][j-1][k-1] - f_i*go,  /* no gap -> gap x */
					  Ixy    [i][j-1][k-1] - f_i*go,  /* gap z  -> gap x */
					  Ixz    [i][j-1][k-1] - f_i*go,  /* gap y  -> gap x */
					  Iyz    [i][j-1][k-1] - f_i*ge,  /* gap x  -> gap x */
					  Ix     [i][j-1][k-1] - f_i*go,  /* gap yz -> gap x */
					  Iy     [i][j-1][k-1] - f_i*ge,  /* gap xz -> gap x */
					  Iz     [i][j-1][k-1] - f_i*ge)  /* gap xy -> gap x */
		    + Syz[s2+j-1][s3+k-1];

		  /* Ix, Iy, Iz */
		  /* gaps in y and z */
		  Ix[i][j][k]=max7float (DPscore[i-1][j][k] - f_j*go-f_k*go,  /* no gap -> gap yz */
					 Ixy    [i-1][j][k] - f_j*go-f_k*ge,  /* gap z  -> gap yz */
					 Ixz    [i-1][j][k] - f_j*ge-f_k*go,  /* gap y  -> gap yz */
					 Iyz    [i-1][j][k] - f_j*go-f_k*go,  /* gap x  -> gap yz */
					 Ix     [i-1][j][k] - f_j*ge-f_k*ge,  /* gap yz -> gap yz */
					 Iy     [i-1][j][k] - f_j*go-f_k*ge,  /* gap xz -> gap yz */
					 Iz     [i-1][j][k] - f_j*ge-f_k*go); /* gap xy -> gap yz */

		  /* gaps in x and z */
		  Iy[i][j][k]=max7float (DPscore[i][j-1][k] - f_i*go-f_k*go,  /* no gap -> gap xz */
					 Ixy    [i][j-1][k] - f_i*go-f_k*ge,  /* gap z  -> gap xz */
					 Ixz    [i][j-1][k] - f_i*go-f_k*go,  /* gap y  -> gap xz */
					 Iyz    [i][j-1][k] - f_i*ge-f_k*go,  /* gap x  -> gap xz */
					 Ix     [i][j-1][k] - f_i*go-f_k*ge,  /* gap yz -> gap xz */
					 Iy     [i][j-1][k] - f_i*ge-f_k*ge,  /* gap xz -> gap xz */
					 Iz     [i][j-1][k] - f_i*ge-f_k*go); /* gap xy -> gap xz */

		  /* gaps in x and y */
		  Iz[i][j][k]=max7float (DPscore[i][j][k-1] - f_i*go-f_j*go,  /* no gap -> gap xy */
					 Ixy    [i][j][k-1] - f_i*go-f_j*go,  /* gap z  -> gap xy */
					 Ixz    [i][j][k-1] - f_i*go-f_j*ge,  /* gap y  -> gap xy */
					 Iyz    [i][j][k-1] - f_i*ge-f_j*go,  /* gap x  -> gap xy */
					 Ix     [i][j][k-1] - f_i*go-f_j*ge,  /* gap yz -> gap xy */
					 Iy     [i][j][k-1] - f_i*ge-f_j*go,  /* gap xz -> gap xy */
					 Iz     [i][j][k-1] - f_i*ge-f_j*ge); /* gap xy -> gap xy */
#endif
		}
	      /* find maximum */
	      maxscorearg=maxarg7float (DPscore[i][j][k], Ixy[i][j][k], Ixz[i][j][k], Iyz[i][j][k], Ix[i][j][k], Iy[i][j][k], Iz[i][j][k]);
	      /* diagonal (align: x-1, y-1, z-1) -> aligned, no gaps */
	      if (maxscorearg==0) 
		{
		  /* DPscore[i][j][k]=DPscore[i][j][k]; */
		  DPdir[i][j][k]=DIRXYZ;
		}
	      /* (align: x-1, y-1) -> not aligned, gap in third sequence */
	      else if (maxscorearg==1)
		{
		  DPscore[i][j][k]=Ixy[i][j][k];
		  DPdir[i][j][k]=DIRXY;
		}
	      /* (align: x-1, z-1) -> not aligned, gap in second sequence */
	      else if (maxscorearg==2)
		{
		  DPscore[i][j][k]=Ixz[i][j][k];
		  DPdir[i][j][k]=DIRXZ;
		}
	      /* (align: y-1, z-1) -> not aligned, gap in first sequence */
	      else if (maxscorearg==3)
		{
		  DPscore[i][j][k]=Iyz[i][j][k];
		  DPdir[i][j][k]=DIRYZ;
		}
	      /* (align: x-1) -> not aligned, gap in second and third sequence */
	      else if (maxscorearg==4)
		{
		  DPscore[i][j][k]=Ix[i][j][k];
		  DPdir[i][j][k]=DIRX;
		}
	      /* (align: y-1) -> not aligned, gap in first and third sequence */
	      else if (maxscorearg==5)
		{
		  DPscore[i][j][k]=Iy[i][j][k];
		  DPdir[i][j][k]=DIRY;
		}
	      /* (align: z-1) -> not aligned, gap in second and third sequence */
	      else
		{
		  DPscore[i][j][k]=Iz[i][j][k];
		  DPdir[i][j][k]=DIRZ;
		}
	    }
	}
    }

  /* traceback */
  /************/
  /* allocate memory for new descriptors */
  MALLOC (newdes, char*, num);
  j=0;
  for (i=0; i<n1; i++, j++)
    {
      len=(int)strlen (ali1->descriptors[i]);
      MALLOC (newdes[j], char, len+1);
      strncpy (newdes[j], ali1->descriptors[i], (size_t)(len+1));
    }
  for (i=0; i<n2; i++, j++)
    {
      len=(int)strlen (ali2->descriptors[i]);
      MALLOC (newdes[j], char, len+1);
      strncpy (newdes[j], ali2->descriptors[i], (size_t)(len+1));
    }
  for (i=0; i<n3; i++, j++)
    {
      len=(int)strlen (ali3->descriptors[i]);
      MALLOC (newdes[j], char, len+1);
      strncpy (newdes[j], ali3->descriptors[i], (size_t)(len+1));
    }

  /* allocate memory for new sequences */
  MALLOC (newseqs, char*, num);
  for (i=0; i<num; i++) MALLOC (newseqs[i], char, msum);

  /* allocate memory for new structural information */
  MALLOC (newsinfo, SINFO*, num);
  for (i=0; i<num; i++) MALLOC (newsinfo[i], SINFO, msum);

  /* start traceback */
  len=0;
  posi=l1;
  posj=l2;
  posk=l3;

  /* build the strings and the structural information in reverse order */
  while (DPdir[posi][posj][posk]!=STOP)
    {
      if (DPdir[posi][posj][posk]==DIRXYZ)
	{
	  j=0; posi--; posj--; posk--;
	  for (i=0; i<n1; i++, j++)
	    {
	      newseqs [j][len]=ali1->sequences[i][s1+posi]; 
	      newsinfo[j][len]=ali1->sinfo    [i][s1+posi];
	    }
	  for (i=0; i<n2; i++, j++)
	    {
	      newseqs [j][len]=ali2->sequences[i][s2+posj]; 
	      newsinfo[j][len]=ali2->sinfo    [i][s2+posj];
	    }
	  for (i=0; i<n3; i++, j++) 
	    {
	      newseqs [j][len]=ali3->sequences[i][s3+posk]; 
	      newsinfo[j][len]=ali3->sinfo    [i][s3+posk];
	    }
	}
      else if (DPdir[posi][posj][posk]==DIRXY)
	{
	  j=0; posi--; posj--;
	  for (i=0; i<n1; i++, j++) 
	    {
	      newseqs [j][len]=ali1->sequences[i][s1+posi]; 
	      newsinfo[j][len]=ali1->sinfo    [i][s1+posi];
	    }
	  for (i=0; i<n2; i++, j++)
	    {
	      newseqs [j][len]=ali2->sequences[i][s2+posj]; 
	      newsinfo[j][len]=ali2->sinfo    [i][s2+posj];
	    }
	  for (i=0; i<n3; i++, j++)
	    {
	      newseqs [j][len]=GAPCHAR; 
	      newsinfo[j][len]=SINFOZERO;
	    }	      
	}
      else if (DPdir[posi][posj][posk]==DIRXZ)
	{
	  j=0; posi--; posk--;
	  for (i=0; i<n1; i++, j++) 
	    {
	      newseqs [j][len]=ali1->sequences[i][s1+posi]; 
	      newsinfo[j][len]=ali1->sinfo    [i][s1+posi];
	    }
	  for (i=0; i<n2; i++, j++)
	    {
	      newseqs [j][len]=GAPCHAR; 
	      newsinfo[j][len]=SINFOZERO;
	    }
	  for (i=0; i<n3; i++, j++)
	    {
	      newseqs [j][len]=ali3->sequences[i][s3+posk]; 
	      newsinfo[j][len]=ali3->sinfo    [i][s3+posk];
	    }
	}
      else if (DPdir[posi][posj][posk]==DIRYZ)
	{
	  j=0; posj--; posk--;
	  for (i=0; i<n1; i++, j++) 
	    {
	      newseqs [j][len]=GAPCHAR; 
	      newsinfo[j][len]=SINFOZERO;
	    }
	  for (i=0; i<n2; i++, j++)
	    {
	      newseqs [j][len]=ali2->sequences[i][s2+posj]; 
	      newsinfo[j][len]=ali2->sinfo    [i][s2+posj];
	    }
	  for (i=0; i<n3; i++, j++)
	    {
	      newseqs [j][len]=ali3->sequences[i][s3+posk];
	      newsinfo[j][len]=ali3->sinfo    [i][s3+posk];
	    }
	}
      else if (DPdir[posi][posj][posk]==DIRX)
	{
	  j=0; posi--;
	  for (i=0; i<n1; i++, j++) 
	    {
	      newseqs [j][len]=ali1->sequences[i][s1+posi]; 
	      newsinfo[j][len]=ali1->sinfo    [i][s1+posi];
	    }
	  for (i=0; i<n2; i++, j++) 
	    {
	      newseqs [j][len]=GAPCHAR; 
	      newsinfo[j][len]=SINFOZERO;
	    }
	  for (i=0; i<n3; i++, j++)
	    {
	      newseqs [j][len]=GAPCHAR; 
	      newsinfo[j][len]=SINFOZERO;
	    }
	}
      else if (DPdir[posi][posj][posk]==DIRY)
	{
	  j=0; posj--;
	  for (i=0; i<n1; i++, j++) 
	    {
	      newseqs [j][len]=GAPCHAR;
	      newsinfo[j][len]=SINFOZERO;
	    }
	  for (i=0; i<n2; i++, j++) 
	    {
	      newseqs [j][len]=ali2->sequences[i][s2+posj]; 
	      newsinfo[j][len]=ali2->sinfo    [i][s2+posj];
	    }
	  for (i=0; i<n3; i++, j++) 
	    {
	      newseqs [j][len]=GAPCHAR; 
	      newsinfo[j][len]=SINFOZERO;
	    }
	}
      else if (DPdir[posi][posj][posk]==DIRZ)
	{
	  j=0; posk--;
	  for (i=0; i<n1; i++, j++) 
	    {
	      newseqs [j][len]=GAPCHAR;
	      newsinfo[j][len]=SINFOZERO;
	    }
	  for (i=0; i<n2; i++, j++)
	    {
	      newseqs [j][len]=GAPCHAR;
	      newsinfo[j][len]=SINFOZERO;
	    }
	  for (i=0; i<n3; i++, j++) 
	    {
	      newseqs [j][len]=ali3->sequences[i][s3+posk]; 
	      newsinfo[j][len]=ali3->sinfo    [i][s3+posk];
	    }
	}
      else FATAL ("Unknown 3D traceback direction.");
      len++;
    }

  /* terminate and reverse the strings */
  for (i=0; i<num; i++)
    {
      REALLOC (newseqs[i], char, len+1);
      newseqs[i][len]=EOS;
      stringReverse (newseqs[i]);
    }

  /* terminate and reverse structural information */
  for (i=0; i<num; i++)
    {
      REALLOC (newsinfo[i], SINFO, len);
      sInfoReverse (newsinfo[i], len);
    }

  /* make ali record */
  MALLOC (ali, ALI, 1);

  /* store data into the alignment record */
  /* alignment id */
  ali->id=0;
  /* number of alignments */
  ali->num=num;
  /* all descriptors of the alignment */
  ali->descriptors=newdes;
  /* all sequences of the alignment */
  ali->sequences=newseqs;
  /* structural information */
  ali->sinfo=newsinfo;
  /* length of the alignment */
  ali->len=len;
  /* ids of all sequences */
  MALLOC (ali->sids, int, num);
  i=0;
  for (j=0; j<n1; j++, i++) ali->sids[i]=ali1->sids[j];
  for (j=0; j<n2; j++, i++) ali->sids[i]=ali2->sids[j];
  for (j=0; j<n3; j++, i++) ali->sids[i]=ali3->sids[j];

  /* update alignment score */
  ali->score=DPscore[l1][l2][l3];

  /* free memory */
  /***************/
  FREE3D (DPscore, m1, m2, i, j);
  FREE3D (Ix, m1, m2, i, j);
  FREE3D (Iy, m1, m2, i, j);
  FREE3D (Iz, m1, m2, i, j);
  FREE3D (Ixy, m1, m2, i, j);
  FREE3D (Ixz, m1, m2, i, j);
  FREE3D (Iyz, m1, m2, i, j);
  FREE3D (DPdir, m1, m2, i, j);

  return (ali);
}
