#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <getopt.h>

#include "global.h"
#include "seqio.h"
#include "options.h"
#include "macros.h"
#include "cmdline.h"
#include "matrices.h"

#if HAVE_CONFIG_H
#include "../config.h"
#endif

extern SEQ Seqs;

static struct option long_options[]={
  {"matrix"        , required_argument, 0, 0  },
  {"match"         , required_argument, 0, 0  },
  {"mismatch"      , required_argument, 0, 0  },
  {"go"            , required_argument, 0, 0  },
  {"ge"            , required_argument, 0, 0  },
  {"Dmatrix"       , required_argument, 0, 0  },
  {"Dmatch"        , required_argument, 0, 0  },
  {"Dmismatch"     , required_argument, 0, 0  },
  {"Dgo"           , required_argument, 0, 0  },
  {"Dge"           , required_argument, 0, 0  },
  {"Amatrix"       , required_argument, 0, 0  },
  {"Amatch"        , required_argument, 0, 0  },
  {"Amismatch"     , required_argument, 0, 0  },
  {"Ago"           , required_argument, 0, 0  },
  {"Age"           , required_argument, 0, 0  },
  {"trans"         , required_argument, 0, 0  },
  {"wobble"        , required_argument, 0, 0  },
  {"psi"           , required_argument, 0, 0  },
  {"seqtype"       , required_argument, 0, 0  },
  {"distance"      , required_argument, 0, 0  },
  {"score"         , required_argument, 0, 0  },
  {"shuffles"      , required_argument, 0, 0  },
  {"dclength"      , required_argument, 0, 0  },
  {"weightseqs"    , no_argument,       0, 'w'},
  {"noweightseqs"  , no_argument,       0, 'W'},
  {"weightterms"   , no_argument,       0, 't'},
  {"simplegap"     , no_argument,       0, 'G'},
  {"extendedgap"   , no_argument,       0, 'g'},
  {"noweightterms" , no_argument,       0, 'T'},
  {"structure"     , no_argument,       0, 'u'},
  {"nostructure"   , no_argument,       0, 'U'},
  {"sort"          , no_argument,       0, 's'},
  {"nosort"        , no_argument,       0, 'S'},
  {"clustal"       , no_argument,       0, 'C'},
  {"msf"           , no_argument,       0, 'M'},
  {"verbose"       , no_argument,       0, 'v'},
  {"quiet"         , no_argument,       0, 'q'},
  {"help"          , no_argument,       0, 'h'},
  {0               , 0,                 0,   0}
};

extern OPTION Options;

/*********************************************************************
shows usage / help screen
*********************************************************************/
static void usage (void)
{
  printf ("NAME\n");
  printf ("     %s %s - Multiple Sequence Alignment Tool\n\n", PACKAGE_NAME, PACKAGE_VERSION);
  printf ("SYNOPSIS\n");
  printf ("     %-10s [-wWtTuGgSsUCMLvqh] [--[D|A]match FLOAT] [--[D|A]mismatch FLOAT]\n", PACKAGE_NAME);
  printf ("     %-10s [--trans FLOAT] [--wobble FLOAT] [--[D|A]go FLOAT] [--[D|A]ge FLOAT]\n", "");
  printf ("     %-10s [--psi FLOAT] [--seqtype STRING [--matrix STRIG]\n", "");
  printf ("     %-10s [fastafile] [alignmentfile]\n", "");
  printf ("\n");
  printf ("OPTIONS\n");
  printf ("     --distance           mean pairwise distance calculation of sequences\n");
  printf ("     --score              score calculation of given alignment\n");
  printf ("\n");
  printf ("     --seqtype STRING     sequence type (dna:     DNA sequences %s\n", (ST==SEQ_DNA ? "(default)":""));
  printf ("                                         rna:     RNA sequences %s\n", (ST==SEQ_RNA ? "(default)":""));
  printf ("                                         protein: protein sequences %s\n", (ST==SEQ_PROTEIN ? "(default)":""));
  printf ("                                         auto:    auto-detect %s)\n", (ST==SEQ_AUTO ? "(default)":""));
  printf ("\n");
  printf ("     The following options consider distance calculation paramenters:\n");
  printf ("     --Dmatrix STRING   substitution matrix for\n");
  printf ("                        nucleotide sequences:\n");
  printf ("                            - MAMMA (Match/Mismatch (default))\n");
  printf ("                            - IUB\n");
  printf ("                        amino acid sequences:\n");
  printf ("                            - MAMMA (Match/MisMatch)\n");
  printf ("                            - BLOSUM, BLOSUM30, BLOSUM40, BLOSUM45, BLOSUM62 (default), BLOSUM80\n");
  printf ("                            - PAM, PAM20, PAM60, PAM120, PAM250, PAM350\n");
  printf ("                            - GONNET, GONNET40, GONNET80, GONNET120, GONNET160,\n");
  printf ("                              GONNET250, GONNET300, GONNET350\n");
  printf ("     --Dmatch FLOAT     score for matching basepairs (default: %2.1f)\n", D_MATCH);
  printf ("     --Dmismatch FLOAT  score for mis-matching base pairs (default: %2.1f)\n", D_MISMATCH);
  printf ("     --Dgo FLOAT        gap-open penalty (default: %2.1f for DNA\n", D_NUCGAPOPEN);
  printf ("                                                   %2.1f for protein)\n", D_AAGAPOPEN);
  printf ("     --Dge FLOAT        gap-extension penalty (default: %2.1f for DNA\n", D_NUCGAPEXTEND);
  printf ("                                                        %2.1f for protein)\n", D_AAGAPEXTEND);
  printf ("     --shuffles INT     number of sequence shuffles for approximate distance calculation (default: %d)\n", NUMSHUFFLES);
  printf ("                        (a value of zero uses fast but approximative distance calculation)\n");
  printf ("\n");
  printf ("     The following options consider alignment paramenters:\n");
  printf ("     --Amatrix STRING     substitution matrix for\n");
  printf ("                          nucleotide sequences:\n");
  printf ("                            - MAMMA (Match/Mismatch (default))\n");
  printf ("                            - IUB\n");
  printf ("                          amino acid sequences:\n");
  printf ("                            - MAMMA (Match/MisMatch)\n");
  printf ("                            - BLOSUM, BLOSUM30, BLOSUM40, BLOSUM45, BLOSUM62 (default),\n");
  printf ("                              BLOSUM80\n");
  printf ("                            - PAM, PAM20, PAM60, PAM120, PAM250, PAM350\n");
  printf ("                            - GONNET, GONNET40, GONNET80, GONNET120, GONNET160,\n");
  printf ("                              GONNET250, GONNET300, GONNET350\n");
  printf ("                          if no BLOSUM, PAM or GONNET matrix is specified\n");
  printf ("                          explicitely, an appropriate matrix is choosen\n");
  printf ("                          automatically\n");
  printf ("     --Amatch FLOAT       score for matching basepairs (default: %2.1f)\n", A_MATCH);
  printf ("     --Amismatch FLOAT    score for mis-matching base pairs (default: %2.1f)\n", A_MISMATCH);
  printf ("     --Ago FLOAT          gap-open penalty (default: %2.1f for DNA\n", A_NUCGAPOPEN);
  printf ("                                                     %2.1f for protein)\n", A_AAGAPOPEN);
  printf ("     --Age FLOAT          gap-extension penalty (default: %2.1f for DNA\n", A_NUCGAPEXTEND);
  printf ("                                                          %2.1f for protein)\n", A_AAGAPEXTEND);
  printf ("\n");
  printf ("                          If the preceding character 'A' or 'D' is omitted, the arguments specified for\n"); 
  printf ("                          the following switches --matrix, --match, --mismatch, --go, and --ge apply\n");
  printf ("                          for both distance and alignment calculation.\n");
  printf ("\n");
  printf ("     --trans FLOAT        transition score factor in range [0.0,1.0]\n");
  printf ("                          (default: %2.1f)\n", DNATRANSIT);
  printf ("     --wobble FLOAT       RNA wobble pair score factor in range [0.0,1.0]\n");
  printf ("                          (default: %2.1f)\n", RNAWOBBLE);
  printf ("     --psi FLOAT          balance between sequence score and structure score\n");
  printf ("                          [0.0-1.0] (default: %2.1f)\n", PSI);
  printf ("     --dclength           divide&conquer length threshold (default: disabled)\n");
  printf ("\n");
  printf ("     -W  --noweightseqs   do not weight sequences %s\n", (SEQWEIGHT==0 ? "(default)":""));
  printf ("     -w  --weightseqs     weight sequences %s\n", (SEQWEIGHT==1 ? "(default)":""));
  printf ("     -T  --noweightterms  do not weight terminal gaps %s\n", (TERMGAPWEIGHT==0 ? "(default)":""));
  printf ("     -t  --weightterms    weight terminal gaps %s\n", (TERMGAPWEIGHT==1 ? "(default)":""));
  printf ("     -G  --simplegap      simple gap costs %s\n", (QUASINATURALGAPS==1 ? "(default)":""));
  printf ("     -g  --extendedgap    extended gap costs %s\n", (QUASINATURALGAPS==0 ? "(default)":""));
  printf ("     -U  --nostructure    do not use structural information\n");
  printf ("     -u  --structure      use structural information\n");
  printf ("     -C  --clustal        use CLUSTAL alignment format for output %s\n", (OUTPUT==OUT_CLUSTAL ? "(default)":""));
  printf ("     -M  --msf            use MSF alignment format for alignment output instead\n");
  printf ("                          of CLUSTAL format %s\n", (OUTPUT==OUT_MSF ? "(default)":""));
  printf ("     -S  --nosort         do not sort alignment to input sequence order\n");
  printf ("     -s  --sort           sort alignment to input sequence order\n");
  printf ("     -v  --verbose        increase verbosity level (standard level: %d)\n", VERBOSITY);
  printf ("     -q  --quiet          quiet mode, no output except errors and warnings is\n");
  printf ("                          generated (=verb. level 0)\n");
  printf ("     -h  --help           this help screen\n");
  printf ("\n");
  printf ("AUTHOR\n");
  printf ("     Written by %s.\n\n", PACKAGE_AUTHOR);
  printf ("REPORTING BUGS\n");
  printf ("     Report bugs to %s.\n\n", PACKAGE_BUGREPORT);
  printf ("COPYRIGHT\n");
  printf ("     This is free software; see the source for copying conditions. There is\n");
  printf ("     NO warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR\n");
  printf ("     PURPOSE\n\n");

  exit (EXIT_SUCCESS);
}

/******************************************************************
parses command line
******************************************************************/
void parseCmdLine (int argc, char **argv)
{
  int c;
  int this_option_optind;
  int option_index;
  float tmp;
  int tmpi;

  while (1)
    {
      this_option_optind=optind ? optind:1;
      option_index=0;

      c=getopt_long (argc, argv, "dwWtTuUgGCMvqh", long_options, &option_index);
      if (c==-1) break;
 
      switch (c)
	{
	case 0: /* long options */
	  if (strcasecmp (long_options[option_index].name, "matrix")==0)
	    {
	      if      (strcasecmp (optarg, "mamma")==0) Options.D_Mattype=MAT_MAMMA;
	      else if (strcasecmp (optarg, "iub")==0) Options.D_Mattype=MAT_IUB;
	      else if (strcasecmp (optarg, "blosum")==0) Options.D_Mattype=MAT_BLOSUM;
	      else if (strcasecmp (optarg, "blosum30")==0) Options.D_Mattype=MAT_BLOSUM30;
	      else if (strcasecmp (optarg, "blosum40")==0) Options.D_Mattype=MAT_BLOSUM40;
	      else if (strcasecmp (optarg, "blosum45")==0) Options.D_Mattype=MAT_BLOSUM45;
	      else if (strcasecmp (optarg, "blosum62")==0) Options.D_Mattype=MAT_BLOSUM62;
	      else if (strcasecmp (optarg, "blosum80")==0) Options.D_Mattype=MAT_BLOSUM80;
	      else if (strcasecmp (optarg, "pam")==0) Options.D_Mattype=MAT_PAM;
	      else if (strcasecmp (optarg, "pam20")==0) Options.D_Mattype=MAT_PAM20;
	      else if (strcasecmp (optarg, "pam60")==0) Options.D_Mattype=MAT_PAM60;
	      else if (strcasecmp (optarg, "pam120")==0) Options.D_Mattype=MAT_PAM120;
	      else if (strcasecmp (optarg, "pam250")==0) Options.D_Mattype=MAT_PAM250;
	      else if (strcasecmp (optarg, "pam350")==0) Options.D_Mattype=MAT_PAM350;
	      else if (strcasecmp (optarg, "gonnet")==0) Options.D_Mattype=MAT_GONNET;
	      else if (strcasecmp (optarg, "gonnet40")==0) Options.D_Mattype=MAT_GONNET40;
	      else if (strcasecmp (optarg, "gonnet80")==0) Options.D_Mattype=MAT_GONNET80;
	      else if (strcasecmp (optarg, "gonnet120")==0) Options.D_Mattype=MAT_GONNET120;
	      else if (strcasecmp (optarg, "gonnet160")==0) Options.D_Mattype=MAT_GONNET160;
	      else if (strcasecmp (optarg, "gonnet250")==0) Options.D_Mattype=MAT_GONNET250;
	      else if (strcasecmp (optarg, "gonnet300")==0) Options.D_Mattype=MAT_GONNET300;
	      else if (strcasecmp (optarg, "gonnet350")==0) Options.D_Mattype=MAT_GONNET350;
	      else FATAL ("follow -mattype with valid matrix type");
	      if      (strcasecmp (optarg, "mamma")==0) Options.A_Mattype=MAT_MAMMA;
	      else if (strcasecmp (optarg, "iub")==0) Options.A_Mattype=MAT_IUB;
	      else if (strcasecmp (optarg, "blosum")==0) Options.A_Mattype=MAT_BLOSUM;
	      else if (strcasecmp (optarg, "blosum30")==0) Options.A_Mattype=MAT_BLOSUM30;
	      else if (strcasecmp (optarg, "blosum40")==0) Options.A_Mattype=MAT_BLOSUM40;
	      else if (strcasecmp (optarg, "blosum45")==0) Options.A_Mattype=MAT_BLOSUM45;
	      else if (strcasecmp (optarg, "blosum62")==0) Options.A_Mattype=MAT_BLOSUM62;
	      else if (strcasecmp (optarg, "blosum80")==0) Options.A_Mattype=MAT_BLOSUM80;
	      else if (strcasecmp (optarg, "pam")==0) Options.A_Mattype=MAT_PAM;
	      else if (strcasecmp (optarg, "pam20")==0) Options.A_Mattype=MAT_PAM20;
	      else if (strcasecmp (optarg, "pam60")==0) Options.A_Mattype=MAT_PAM60;
	      else if (strcasecmp (optarg, "pam120")==0) Options.A_Mattype=MAT_PAM120;
	      else if (strcasecmp (optarg, "pam250")==0) Options.A_Mattype=MAT_PAM250;
	      else if (strcasecmp (optarg, "pam350")==0) Options.A_Mattype=MAT_PAM350;
	      else if (strcasecmp (optarg, "gonnet")==0) Options.A_Mattype=MAT_GONNET;
	      else if (strcasecmp (optarg, "gonnet40")==0) Options.A_Mattype=MAT_GONNET40;
	      else if (strcasecmp (optarg, "gonnet80")==0) Options.A_Mattype=MAT_GONNET80;
	      else if (strcasecmp (optarg, "gonnet120")==0) Options.A_Mattype=MAT_GONNET120;
	      else if (strcasecmp (optarg, "gonnet160")==0) Options.A_Mattype=MAT_GONNET160;
	      else if (strcasecmp (optarg, "gonnet250")==0) Options.A_Mattype=MAT_GONNET250;
	      else if (strcasecmp (optarg, "gonnet300")==0) Options.A_Mattype=MAT_GONNET300;
	      else if (strcasecmp (optarg, "gonnet350")==0) Options.A_Mattype=MAT_GONNET350;
	      else FATAL ("follow -mattype with valid matrix type");
	      if (((Options.D_Matchgiven)    || (Options.A_Matchgiven) ||
		   (Options.D_Mismatchgiven) || (Options.A_Mismatchgiven)) &&
		  ((Options.D_Mattype!=MAT_MAMMA) || (Options.A_Mattype!=MAT_MAMMA)))
		WARN ("Provided scores do not apply for this matrix type.");
	    }
	  else if (strcasecmp (long_options[option_index].name, "Dmatrix")==0)
	    {
	      if      (strcasecmp (optarg, "mamma")==0) Options.D_Mattype=MAT_MAMMA;
	      else if (strcasecmp (optarg, "iub")==0) Options.D_Mattype=MAT_IUB;
	      else if (strcasecmp (optarg, "blosum")==0) Options.D_Mattype=MAT_BLOSUM;
	      else if (strcasecmp (optarg, "blosum30")==0) Options.D_Mattype=MAT_BLOSUM30;
	      else if (strcasecmp (optarg, "blosum40")==0) Options.D_Mattype=MAT_BLOSUM40;
	      else if (strcasecmp (optarg, "blosum45")==0) Options.D_Mattype=MAT_BLOSUM45;
	      else if (strcasecmp (optarg, "blosum62")==0) Options.D_Mattype=MAT_BLOSUM62;
	      else if (strcasecmp (optarg, "blosum80")==0) Options.D_Mattype=MAT_BLOSUM80;
	      else if (strcasecmp (optarg, "pam")==0) Options.D_Mattype=MAT_PAM;
	      else if (strcasecmp (optarg, "pam20")==0) Options.D_Mattype=MAT_PAM20;
	      else if (strcasecmp (optarg, "pam60")==0) Options.D_Mattype=MAT_PAM60;
	      else if (strcasecmp (optarg, "pam120")==0) Options.D_Mattype=MAT_PAM120;
	      else if (strcasecmp (optarg, "pam250")==0) Options.D_Mattype=MAT_PAM250;
	      else if (strcasecmp (optarg, "pam350")==0) Options.D_Mattype=MAT_PAM350;
	      else if (strcasecmp (optarg, "gonnet")==0) Options.D_Mattype=MAT_GONNET;
	      else if (strcasecmp (optarg, "gonnet40")==0) Options.D_Mattype=MAT_GONNET40;
	      else if (strcasecmp (optarg, "gonnet80")==0) Options.D_Mattype=MAT_GONNET80;
	      else if (strcasecmp (optarg, "gonnet120")==0) Options.D_Mattype=MAT_GONNET120;
	      else if (strcasecmp (optarg, "gonnet160")==0) Options.D_Mattype=MAT_GONNET160;
	      else if (strcasecmp (optarg, "gonnet250")==0) Options.D_Mattype=MAT_GONNET250;
	      else if (strcasecmp (optarg, "gonnet300")==0) Options.D_Mattype=MAT_GONNET300;
	      else if (strcasecmp (optarg, "gonnet350")==0) Options.D_Mattype=MAT_GONNET350;
	      else FATAL ("follow -mattype with valid matrix type");
	      if (((Options.D_Matchgiven) || (Options.D_Mismatchgiven)) && (Options.D_Mattype!=MAT_MAMMA))
		WARN ("Provided scores do not apply for this matrix type.");
	    }
	  else if (strcasecmp (long_options[option_index].name, "Amatrix")==0)
	    {
	      if      (strcasecmp (optarg, "mamma")==0) Options.A_Mattype=MAT_MAMMA;
	      else if (strcasecmp (optarg, "iub")==0) Options.A_Mattype=MAT_IUB;
	      else if (strcasecmp (optarg, "blosum")==0) Options.A_Mattype=MAT_BLOSUM;
	      else if (strcasecmp (optarg, "blosum30")==0) Options.A_Mattype=MAT_BLOSUM30;
	      else if (strcasecmp (optarg, "blosum40")==0) Options.A_Mattype=MAT_BLOSUM40;
	      else if (strcasecmp (optarg, "blosum45")==0) Options.A_Mattype=MAT_BLOSUM45;
	      else if (strcasecmp (optarg, "blosum62")==0) Options.A_Mattype=MAT_BLOSUM62;
	      else if (strcasecmp (optarg, "blosum80")==0) Options.A_Mattype=MAT_BLOSUM80;
	      else if (strcasecmp (optarg, "pam")==0) Options.A_Mattype=MAT_PAM;
	      else if (strcasecmp (optarg, "pam20")==0) Options.A_Mattype=MAT_PAM20;
	      else if (strcasecmp (optarg, "pam60")==0) Options.A_Mattype=MAT_PAM60;
	      else if (strcasecmp (optarg, "pam120")==0) Options.A_Mattype=MAT_PAM120;
	      else if (strcasecmp (optarg, "pam250")==0) Options.A_Mattype=MAT_PAM250;
	      else if (strcasecmp (optarg, "pam350")==0) Options.A_Mattype=MAT_PAM350;
	      else if (strcasecmp (optarg, "gonnet")==0) Options.A_Mattype=MAT_GONNET;
	      else if (strcasecmp (optarg, "gonnet40")==0) Options.A_Mattype=MAT_GONNET40;
	      else if (strcasecmp (optarg, "gonnet80")==0) Options.A_Mattype=MAT_GONNET80;
	      else if (strcasecmp (optarg, "gonnet120")==0) Options.A_Mattype=MAT_GONNET120;
	      else if (strcasecmp (optarg, "gonnet160")==0) Options.A_Mattype=MAT_GONNET160;
	      else if (strcasecmp (optarg, "gonnet250")==0) Options.A_Mattype=MAT_GONNET250;
	      else if (strcasecmp (optarg, "gonnet300")==0) Options.A_Mattype=MAT_GONNET300;
	      else if (strcasecmp (optarg, "gonnet350")==0) Options.A_Mattype=MAT_GONNET350;
	      else FATAL ("follow -mattype with valid matrix type");
	      if (((Options.A_Matchgiven) || (Options.A_Mismatchgiven)) && (Options.A_Mattype!=MAT_MAMMA))
		WARN ("Provided scores do not apply for this matrix type.");
	    }
	  else if (strcasecmp (long_options[option_index].name, "match")==0)
	    {
	      if (sscanf (optarg, "%f", &tmp)==0) { FATAL ("Option --match requires argument."); }
	      else
		{ 
		  Options.D_Match=tmp;
		  Options.A_Match=tmp;
		  Options.D_Matchgiven=TRUE;
		  Options.A_Matchgiven=TRUE;
		  if (Options.D_Mattype==MAT_DEFAULT) Options.D_Mattype=MAT_MAMMA;
		  if (Options.A_Mattype==MAT_DEFAULT) Options.A_Mattype=MAT_MAMMA;
		  if ((Options.D_Mattype!=MAT_MAMMA) || (Options.A_Mattype!=MAT_MAMMA)) 
		    WARN ("Provided match score only apply for MAMMA matrix type.");
		}
	    }
	  else if (strcasecmp (long_options[option_index].name, "Dmatch")==0)
	    {
	      if (sscanf (optarg, "%f", &tmp)==0) { FATAL ("Option --Dmatch requires argument."); }
	      else
		{ 
		  Options.D_Match=tmp;
		  Options.D_Matchgiven=TRUE;
		  if (Options.D_Mattype==MAT_DEFAULT) Options.D_Mattype=MAT_MAMMA;
		  if (Options.D_Mattype!=MAT_MAMMA) WARN ("Provided match score only apply for MAMMA matrix type.");
		}
	    }
	  else if (strcasecmp (long_options[option_index].name, "Amatch")==0)
	    {
	      if (sscanf (optarg, "%f", &tmp)==0) { FATAL ("Option --Amatch requires argument."); }
	      else
		{ 
		  Options.A_Match=tmp;
		  Options.A_Matchgiven=TRUE;
		  if (Options.A_Mattype==MAT_DEFAULT) Options.A_Mattype=MAT_MAMMA;
		  if (Options.A_Mattype!=MAT_MAMMA) WARN ("Provided match score only apply for MAMMA matrix type.");
		}
	    }
	  else if (strcasecmp (long_options[option_index].name, "mismatch")==0)
	    {
	      if (sscanf (optarg, "%f", &tmp)==0) { FATAL ("Option --mismatch requires argument."); }
	      else
		{
		  Options.D_Mismatch=tmp;
		  Options.A_Mismatch=tmp;
		  Options.D_Mismatchgiven=TRUE;
		  Options.A_Mismatchgiven=TRUE;
		  if (Options.D_Mattype==MAT_DEFAULT) Options.D_Mattype=MAT_MAMMA;
		  if (Options.A_Mattype==MAT_DEFAULT) Options.A_Mattype=MAT_MAMMA;
		  if ((Options.D_Mattype!=MAT_MAMMA) || (Options.A_Mattype!=MAT_MAMMA)) 
		    WARN ("Provided mis-match score only apply for MAMMA matrix type.");
		}
	    }
	  else if (strcasecmp (long_options[option_index].name, "Dmismatch")==0)
	    {
	      if (sscanf (optarg, "%f", &tmp)==0) { FATAL ("Option --Dmismatch requires argument."); }
	      else
		{
		  Options.D_Mismatch=tmp;
		  Options.D_Mismatchgiven=TRUE;
		  if (Options.D_Mattype==MAT_DEFAULT) Options.D_Mattype=MAT_MAMMA;
		  if (Options.D_Mattype!=MAT_MAMMA) WARN ("Provided mis-match score only apply for MAMMA matrix type.");
		}
	    }
	  else if (strcasecmp (long_options[option_index].name, "Amismatch")==0)
	    {
	      if (sscanf (optarg, "%f", &tmp)==0) { FATAL ("Option --Amismatch requires argument."); }
	      else
		{
		  Options.A_Mismatch=tmp;
		  Options.A_Mismatchgiven=TRUE;
		  if (Options.A_Mattype==MAT_DEFAULT) Options.A_Mattype=MAT_MAMMA;
		  if (Options.A_Mattype!=MAT_MAMMA) WARN ("Provided mis-match score only apply for MAMMA matrix type.");
		}
	    }
	  else if (strcasecmp (long_options[option_index].name, "go")==0)
	    {
	      if (sscanf (optarg, "%f", &tmp)==0) { FATAL ("Option --go requires argument."); }
	      else 
		{
		  Options.A_Gapopen=tmp;
		  Options.D_Gapopen=tmp;
		  Options.A_Gogiven=TRUE;
		  Options.D_Gogiven=TRUE;
		}
	    }
	  else if (strcasecmp (long_options[option_index].name, "Dgo")==0)
	    {
	      if (sscanf (optarg, "%f", &tmp)==0) { FATAL ("Option --Dgo requires argument."); }
	      else 
		{
		  Options.D_Gapopen=tmp;
		  Options.D_Gogiven=TRUE;
		}
	    }
	  else if (strcasecmp (long_options[option_index].name, "Ago")==0)
	    {
	      if (sscanf (optarg, "%f", &tmp)==0) { FATAL ("Option --Ago requires argument."); }
	      else 
		{
		  Options.A_Gapopen=tmp;
		  Options.A_Gogiven=TRUE;
		}
	    }

	  else if (strcasecmp (long_options[option_index].name, "ge")==0)
	    {
	      if (sscanf (optarg, "%f", &tmp)==0) { FATAL ("Option --ge requires argument."); }
	      else 
		{
		  Options.D_Gapextend=tmp;
		  Options.A_Gapextend=tmp;
		  Options.D_Gegiven=TRUE;
		  Options.A_Gegiven=TRUE;
		}
	    }
	  else if (strcasecmp (long_options[option_index].name, "Dge")==0)
	    {
	      if (sscanf (optarg, "%f", &tmp)==0) { FATAL ("Option --Dge requires argument."); }
	      else 
		{
		  Options.D_Gapextend=tmp;
		  Options.D_Gegiven=TRUE;
		}
	    }
	  else if (strcasecmp (long_options[option_index].name, "Age")==0)
	    {
	      if (sscanf (optarg, "%f", &tmp)==0) { FATAL ("Option --Age requires argument."); }
	      else 
		{
		  Options.A_Gapextend=tmp;
		  Options.A_Gegiven=TRUE;
		}
	    }

	  else if (strcasecmp (long_options[option_index].name, "trans")==0) 
	    {
	      if (sscanf (optarg, "%f", &tmp)==0) { FATAL ("Option --trans requires argument."); }
	      else
		{
		  if (tmp<0.0f) tmp=0.0f;
		  else if (tmp>1.0f) tmp=1.0f;
		  Options.Dnatransit=tmp;
		  if (Options.A_Mattype==MAT_DEFAULT) Options.A_Mattype=MAT_MAMMA;
		}
	    }
	  else if (strcasecmp (long_options[option_index].name, "wobble")==0) 
	    {
	      if (sscanf (optarg, "%f", &tmp)==0) { FATAL ("Option --wobble requires argument."); }
	      else
		{
		  if (tmp<0.0f) tmp=0.0f;
		  else if (tmp>1.0f) tmp=1.0f;
		  Options.Rnawobble=tmp;
		  if (Options.A_Mattype==MAT_DEFAULT) Options.A_Mattype=MAT_MAMMA;
		}
	    }
	  else if (strcasecmp (long_options[option_index].name, "psi")==0)
	    {
	      if (sscanf (optarg, "%f", &tmp)==0) { FATAL ("Option --psi requires argument."); }
	      else Options.Psi=tmp;
	    }
	  else if (strcasecmp (long_options[option_index].name, "seqtype")==0)
	    {
	      if (strcasecmp (optarg, "dna")==0) Options.Seqtype=SEQ_DNA;
	      else if (strcasecmp (optarg, "rna")==0) Options.Seqtype=SEQ_RNA;
	      else if (strcasecmp (optarg, "protein")==0) Options.Seqtype=SEQ_PROTEIN;
	      else if (strcasecmp (optarg, "auto")==0) Options.Seqtype=SEQ_AUTO;
	      else FATAL ("follow -seqtype with 'dna', 'rna', 'protein' or 'auto'");
	    }
	  else if (strcasecmp (long_options[option_index].name, "shuffles")==0)
	    {
	      if (sscanf (optarg, "%d", &tmpi)==0) { FATAL ("Option --shuffles requires argument."); }
	      else Options.Numshuffles=tmpi;
	    }
	  else if (strcasecmp (long_options[option_index].name, "distance")==0) Options.CalcDistanceOnly=TRUE;
	  else if (strcasecmp (long_options[option_index].name, "score")==0) Options.CalcScoreOnly=TRUE;
	  else if (strcasecmp (long_options[option_index].name, "weightseqs")==0) Options.Seqweight=TRUE;
	  else if (strcasecmp (long_options[option_index].name, "noweightseqs")==0) Options.Seqweight=FALSE;
	  else if (strcasecmp (long_options[option_index].name, "simplegap")==0) Options.Gaptype=GAP_SIMPLE;
	  else if (strcasecmp (long_options[option_index].name, "extendedgap")==0) Options.Gaptype=GAP_EXTENDED;
	  else if (strcasecmp (long_options[option_index].name, "weightterms")==0) Options.Termgapweight=TRUE;
	  else if (strcasecmp (long_options[option_index].name, "noweightterms")==0) Options.Termgapweight=FALSE;
	  else if (strcasecmp (long_options[option_index].name, "structure")==0) Options.Usestruc=TRUE;
	  else if (strcasecmp (long_options[option_index].name, "nostructure")==0) Options.Usestruc=FALSE;
	  else if (strcasecmp (long_options[option_index].name, "clustal")==0) Options.Outtype=OUT_CLUSTAL;
	  else if (strcasecmp (long_options[option_index].name, "msf")==0) Options.Outtype=OUT_MSF;
	  else if (strcasecmp (long_options[option_index].name, "sort")==0) Options.Sortali=TRUE;
	  else if (strcasecmp (long_options[option_index].name, "nosort")==0) Options.Sortali=FALSE;
	  else if (strcasecmp (long_options[option_index].name, "verbose")==0) Options.Verbosity++;
	  else if (strcasecmp (long_options[option_index].name, "quiet")==0) Options.Verbosity=0;
	  else if (strcasecmp (long_options[option_index].name, "help")==0) usage ();
	  else if (strcasecmp (long_options[option_index].name, "dclength")==0)
	    {
	      if (sscanf (optarg, "%d", &tmpi)==0) { FATAL ("Option --dclength requires argument."); }
	      else
		{
		  Options.DCLength=tmpi;
		}
	    }
	  else { printf ("%s\n", long_options[option_index].name); exit (1); }
	  break;
	case 'w':
	  Options.Seqweight=TRUE;
	  break;
	case 'W':
	  Options.Seqweight=FALSE;
	  break;
	case 'g':
	  Options.Gaptype=GAP_EXTENDED;
	  break;
	case 'G':
	  Options.Gaptype=GAP_SIMPLE;
	  break;
	case 't':
	  Options.Termgapweight=TRUE;
	  break;
	case 'T':
	  Options.Termgapweight=FALSE;
	  break;
	case 'u':
	  Options.Usestruc=TRUE;
	  break;
	case 'U':
	  Options.Usestruc=FALSE;
	  break;
	case 'C':
	  Options.Outtype=OUT_CLUSTAL;
	  break;
	case 'M':
	  Options.Outtype=OUT_MSF;
	  break;
	case 's':
	  Options.Sortali=TRUE;
	  break;
	case 'S':
	  Options.Sortali=FALSE;
	  break;
	case 'v':
	  Options.Verbosity++;
	  break;
	case 'q':
	  Options.Verbosity=0;
	  break;
	case 'h':
	  usage ();
	  break;
	default:
	  FATAL ("Type -h for options.");
	}
    }
  if (optind<argc) strcpy (Options.Infilename, argv[optind++]);
  if (optind<argc) strcpy (Options.Outfilename, argv[optind++]);
   
  /* check parameters */
  if ((Options.CalcDistanceOnly) && (Options.CalcScoreOnly))
    {
      FATAL ("Mean distance and score calculation not possible at once.");
    }
  else if (Options.CalcDistanceOnly)
    {
      if (strlen (Options.Infilename)==0) FATAL ("Specify FASTA file for distance calculation.");
    }
  else if (Options.CalcScoreOnly)
    {
      if (strlen (Options.Infilename)==0) FATAL ("Specify alignment file for alignment score calculation.");
    }
  else
    {
      if ((strlen (Options.Infilename)==0) || (strlen (Options.Outfilename)==0))
	FATAL ("Please specify input file and output file name.");
    }
}

/**********************************************************************
show program instance information
**********************************************************************/
void showInfo (void)
{
  if (Options.Verbosity)
    {
      printf ("Parameter info:\n");
      if (Options.Infilename!="\0")  
	printf ("Input file name:                              %s\n", Options.Infilename);
      if (strlen (Options.Outfilename)) 
	printf ("Alignment aln output file name:               %s\n", Options.Outfilename);
      if (Options.CalcDistanceOnly)  
	printf ("Scoring alignment in file                     %s\n", Options.Infilename);
      if (Options.CalcScoreOnly)     
	printf ("Scoring alignment in file                     %s\n", Options.Infilename);
      if (Options.DCLength)     
	printf ("D&C length threshold is                       %d\n", Options.DCLength);
      printf ("Sequence type:                                %s\n", 
	      (Options.Seqtype==SEQ_DNA ? "DNA" : 
	       (Options.Seqtype==SEQ_RNA ? "RNA" : 
		(Options.Seqtype==SEQ_PROTEIN ? "PROTEIN" : "AUTO-DETECT"))));
      printf ("\n");
      printf ("Parameters for distance calculation:\n");
      printf ("Matrix type:                                  ");
      switch (Options.D_Mattype)
	{
	case MAT_DEFAULT: 
	  printf ("default (not specified)");
	  break;
	case MAT_MAMMA: 
	  printf ("Match/Mismatch");
	  break;
	case MAT_IUB: 
	  printf ("IUB");
	  break;
	case MAT_BLOSUM: 
	  printf ("BLOSUM family");
	  break;
	case MAT_BLOSUM30: 
	  printf ("BLOSUM30");
	  break;
	case MAT_BLOSUM40: 
	  printf ("BLOSUM40");
	  break;
	case MAT_BLOSUM45: 
	  printf ("BLOSUM45");
	  break;
	case MAT_BLOSUM62: 
	  printf ("BLOSUM62");
	  break;
	case MAT_BLOSUM80: 
	  printf ("BLOSUM80");
	  break;
	case MAT_PAM: 
	  printf ("PAM family");
	  break;
	case MAT_PAM20: 
	  printf ("PAM20");
	  break;
	case MAT_PAM60: 
	  printf ("PAM60");
	  break;
	case MAT_PAM120: 
	  printf ("PAM120");
	  break;
	case MAT_PAM250: 
	  printf ("PAM250");
	  break;
	case MAT_PAM350: 
	  printf ("PAM350");
	  break;
	case MAT_GONNET: 
	  printf ("GONNET family");
	  break;
	case MAT_GONNET40: 
	  printf ("GONNET40");
	  break;
	case MAT_GONNET80: 
	  printf ("GONNET80");
	  break;
	case MAT_GONNET120: 
	  printf ("GONNET120");
	  break;
	case MAT_GONNET160: 
	  printf ("GONNET160");
	  break;
	case MAT_GONNET250: 
	  printf ("GONNET250");
	  break;
	case MAT_GONNET300: 
	  printf ("GONNET300");
	  break;
	case MAT_GONNET350: 
	  printf ("GONNET350");
	  break;
	}
      printf ("\n");
      if ((Options.D_Mattype==MAT_DEFAULT) || (Options.D_Mattype==MAT_MAMMA))
	{
	  printf ("Score for matches:                            %+3.1f\n", Options.D_Match);
	  printf ("Score for mis-matches:                        %+3.1f\n", Options.D_Mismatch);
	}
      if (Options.D_Gogiven) printf ("Penalty for opening gaps:                     %+3.1f\n", Options.D_Gapopen);
      else printf ("Penalty for opening gaps:                     %s\n", "default (not specified)");
      if (Options.D_Gegiven) printf ("Penalty for extending gaps:                   %+3.1f\n", Options.D_Gapextend);
      else printf ("Penalty for extending gaps:                   %s\n", "default (not specified)");
      printf ("Distance calculation method:                  %s", (Options.Numshuffles ? "accurate" : "fast"));
      if (Options.Numshuffles) printf ("(%d shuffles)\n", Options.Numshuffles);
      else printf ("\n");
      printf ("\n");
      printf ("Parameters for alignment calculation:\n");
      printf ("Matrix type:                                  ");
      switch (Options.A_Mattype)
	{
	case MAT_DEFAULT: 
	  printf ("default (not specified)");
	  break;
	case MAT_MAMMA: 
	  printf ("Match/Mismatch");
	  break;
	case MAT_IUB: 
	  printf ("IUB");
	  break;
	case MAT_BLOSUM: 
	  printf ("BLOSUM family");
	  break;
	case MAT_BLOSUM30: 
	  printf ("BLOSUM30");
	  break;
	case MAT_BLOSUM40: 
	  printf ("BLOSUM40");
	  break;
	case MAT_BLOSUM45: 
	  printf ("BLOSUM45");
	  break;
	case MAT_BLOSUM62: 
	  printf ("BLOSUM62");
	  break;
	case MAT_BLOSUM80: 
	  printf ("BLOSUM80");
	  break;
	case MAT_PAM: 
	  printf ("PAM family");
	  break;
	case MAT_PAM20: 
	  printf ("PAM20");
	  break;
	case MAT_PAM60: 
	  printf ("PAM60");
	  break;
	case MAT_PAM120: 
	  printf ("PAM120");
	  break;
	case MAT_PAM250: 
	  printf ("PAM250");
	  break;
	case MAT_PAM350: 
	  printf ("PAM350");
	  break;
	case MAT_GONNET: 
	  printf ("GONNET family");
	  break;
	case MAT_GONNET40: 
	  printf ("GONNET40");
	  break;
	case MAT_GONNET80: 
	  printf ("GONNET80");
	  break;
	case MAT_GONNET120: 
	  printf ("GONNET120");
	  break;
	case MAT_GONNET160: 
	  printf ("GONNET160");
	  break;
	case MAT_GONNET250: 
	  printf ("GONNET250");
	  break;
	case MAT_GONNET300: 
	  printf ("GONNET300");
	  break;
	case MAT_GONNET350: 
	  printf ("GONNET350");
	  break;
	}
      printf ("\n");
      if ((Options.A_Mattype==MAT_DEFAULT) || (Options.A_Mattype==MAT_MAMMA))
	{
	  printf ("Score for matches:                            %+3.1f\n", Options.A_Match);
	  printf ("Score for mis-matches:                        %+3.1f\n", Options.A_Mismatch);
	  if ((Options.Seqtype==SEQ_AUTO) || (Options.Seqtype==SEQ_DNA) || (Options.Seqtype==SEQ_RNA))
	    {
	      printf ("Transition score factor                       %3.1f\n", Options.Dnatransit);
	      if ((Options.Seqtype==SEQ_AUTO) || (Options.Seqtype==SEQ_RNA))
		{
		  printf ("RNA Wobble pair score factor                  %3.1f\n", Options.Rnawobble);
		}
	    }
	}
      if (Options.A_Gogiven) printf ("Penalty for opening gaps:                     %+3.1f\n", Options.A_Gapopen);
      else printf ("Penalty for opening gaps:                     %s\n", "default (not specified)");
      if (Options.A_Gegiven) printf ("Penalty for extending gaps:                   %+3.1f\n", Options.A_Gapextend);
      else printf ("Penalty for extending gaps:                   %s\n", "default (not specified)");
      printf ("Gap type                                      %s\n", (Options.Gaptype==GAP_SIMPLE ? "SIMPLE" : "EXTENDED"));
      printf ("Use structural information:                   %s\n", (Options.Usestruc ? "YES" : "NO"));
      printf ("Sequence-Structure balance term               %3.1f\n", Options.Psi);
      printf ("Sequences are weighted:                       %s\n", (Options.Seqweight ? "YES" : "NO"));
      printf ("\n");
      printf ("Terminal gaps are weighted:                   %s\n", (Options.Termgapweight ? "YES" : "NO"));
      printf ("Sort alignment to input order:                %s\n", (Options.Sortali ? "YES" : "NO"));
      printf ("Alignment output type is:                     %s\n", (Options.Outtype==OUT_CLUSTAL ? "CLUSTAL" : "MSF"));
      printf ("\n");
    }
}

/********************************************************************
write parameter settings and command line to info file
********************************************************************/
void saveSettings (int argc, char **argv)
{
  char buffer[PATHLEN];
  FILE *out=NULL;
  int i;

  if (Options.Saveparmset)
    {
      if (Options.Outfilename)
	{
	  sprintf (buffer, "%s.info", Options.Outfilename);
	  out=fopen (buffer, "w");
	}
      
      if (out)
	{
	  fprintf (out, "FEATURE: actual (default)\n");
	  fprintf (out, "TERMGAPWEIGHT: %d (%d)\n", Options.Termgapweight, TERMGAPWEIGHT);
	  fprintf (out, "SEQWEIGHT: %d (%d)\n", Options.Seqweight, SEQWEIGHT);
	  fprintf (out, "VERBOSITY: %d (%d)\n", VERBOSITY, Options.Verbosity);
	  fprintf (out, "D_MATCH: %f (%f)\n", Options.D_Match, D_MATCH);
	  fprintf (out, "D_MISMATCH: %f (%f)\n", Options.D_Mismatch, D_MISMATCH);
	  fprintf (out, "D_GAPOPEN: %f\n", Options.D_Gapopen);
	  fprintf (out, "D_GAPEXTEND: %f\n", Options.D_Gapextend);
	  fprintf (out, "D_MATTYPE: %d (%d)\n", Options.D_Mattype, D_MT);
	  fprintf (out, "A_MATCH: %f (%f)\n", Options.A_Match, A_MATCH);
	  fprintf (out, "A_MISMATCH: %f (%f)\n", Options.A_Mismatch, A_MISMATCH);
	  fprintf (out, "A_GAPOPEN: %f\n", Options.A_Gapopen);
	  fprintf (out, "A_GAPEXTEND: %f\n", Options.A_Gapextend);
	  fprintf (out, "A_MATTYPE: %d (%d)\n", Options.A_Mattype, A_MT);
	  fprintf (out, "DNATRANSIT: %f (%f)\n", Options.Dnatransit, DNATRANSIT);
	  fprintf (out, "WOBBLETRANSIT: %f (%f)\n", Options.Rnawobble, RNAWOBBLE);
	  fprintf (out, "STRUCTURE: %d (%d)\n", Options.Usestruc, USESTRUC);
	  fprintf (out, "PSI: %f (%f)\n", Options.Psi, PSI);
	  fprintf (out, "DCLength: %d\n", Options.DCLength);
	  fprintf (out, "SEQTYPE: %d (%d)\n", Options.Seqtype, ST);
	  fprintf (out, "NUMSCHUFFLES: %d (%d)\n", Options.Numshuffles, NUMSHUFFLES);
	  fprintf (out, "GAPTYPE: %d (%d)\n", Options.Gaptype, GT);
	  fprintf (out, "SORTALI: %d (%d)\n", Options.Sortali, SORTALI);
	  fprintf (out, "OUTTYPE: %d (%d)\n", Options.Outtype, OUTPUT);
	  fprintf (out, "PURGEMODE: %d\n", PURGEMODE);
	  fprintf (out, "SCORESTRAL: %d\n", SCORESTRAL);
	  fprintf (out, "BUFLEN: %d\n", BUFLEN);
	  fprintf (out, "LW: %d\n", LW);
	  fprintf (out, "Program call: ");
	  
	  for (i=0; i<argc; ++i) fprintf (out, "%s ", argv[i]);
	  fprintf (out, "\n");
	  
	  fclose (out);
	}
    }
}
