/*******************************************************
calculates the additional cost matrix which is needed
for the calculation of the optimal slicing positions
for the Divide and Conquer method
*******************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "global.h"
#include "dc.h"
#include "macros.h"

extern OPTION Options;

#if DEBUG==1
extern FILE *debugfile;
#endif

/********************************************************
calculate additional cost matrix
********************************************************/
float **calcaddcost (ALI *ali1, ALI *ali2, int s1, int l1, int s2, int l2, float **S, SCORING *scoring, float *W)
{
  int i, j;
  float **DPf, **Ixf, **Iyf, **DPr, **Ixr, **Iyr, **C;
  float f_i, f_j;
  float score;
  int maxarg;

  /* get data */
  BOOL tgw=Options.Termgapweight;
  float go=scoring->go;
  float ge=scoring->ge;
  int m1=l1+1;
  int m2=l2+1;
  int e1=s1+l1;
  int e2=s2+l2;

  /* allocate memory */
  MAKE2DFLOAT (DPf, m1, m2, i);
  MAKE2DFLOAT (Ixf, m1, m2, i);
  MAKE2DFLOAT (Iyf, m1, m2, i);
  MAKE2DFLOAT (DPr, m1, m2, i);
  MAKE2DFLOAT (Ixr, m1, m2, i);
  MAKE2DFLOAT (Iyr, m1, m2, i);
  MAKE2DFLOAT (C, m1, m2, i); /* additional cost matrix */

  /* calculate forward part */
  /* set boundary conditions */
  DPf[0][0]=0;
  Ixf[0][0]=-LARGE;
  Iyf[0][0]=-LARGE;
  for (i=1; i<m1; i++)
    {
      DPf[i][0]=((!tgw) && (s1==0) ? 0.0f : -go-(float)(i-1)*ge);
      Ixf[i][0]=((!tgw) && (s1==0) ? 0.0f : -go-(float)(i-1)*ge);
      Iyf[i][0]=-LARGE;
    }
  for (j=1; j<m2; j++)
    {
      DPf[0][j]=((!tgw) && (s2==0) ? 0.0f : -go-(float)(j-1)*ge);
      Ixf[0][j]=-LARGE;
      Iyf[0][j]=((!tgw) && (s2==0) ? 0.0f : -go-(float)(j-1)*ge);
    }
  
  /* iterate over first alignment */
  for (i=1; i<m1; i++)
    {
      /* iterate over second alignment */
      for (j=1; j<m2; j++)
	{
	  /* set pre-factors */
	  f_i=((!tgw) && (e1==ali1->len) && (i==l1) ? 0.0f : 1.0f);
	  f_j=((!tgw) && (e2==ali2->len) && (j==l2) ? 0.0f : 1.0f);
	  
	  DPf[i][j]=DPf[i-1][j-1]+S[s1+i-1][s2+j-1]; /* add the three pairwise costs */
	  Ixf[i][j]=MAX (DPf[i-1][j]-f_j*go, Ixf[i-1][j]-f_j*ge);
	  Iyf[i][j]=MAX (DPf[i][j-1]-f_i*go, Iyf[i][j-1]-f_i*ge);
	  maxarg=MAXARG3 (DPf[i][j], Ixf[i][j], Iyf[i][j]);
	  if (maxarg==0) /* DPf[i][j]=DPf[i][j] */;
	  else if (maxarg==1) DPf[i][j]=Ixf[i][j];
	  else DPf[i][j]=Iyf[i][j];
	}
    }
  
  /* calculate reverse part */
  /* set boundary conditions */
  DPr[l1][l2]=0;
  Ixr[l1][l2]=-LARGE;
  Iyr[l1][l2]=-LARGE;
  for (i=l1-1; i>=0; i--)
    {
      DPr[i][0]=((!tgw) && (e1==ali1->len) ? 0.0f : -go-(float)(m1-i)*ge);
      Ixr[i][0]=((!tgw) && (e1==ali1->len) ? 0.0f : -go-(float)(m1-1)*ge);
      Iyr[i][0]=-LARGE;
    }
  for (j=l2-1; j>=0; j--)
    {
      DPr[0][j]=((!tgw) && (e2==ali2->len) ? 0.0f : -go-(float)(m2-j)*ge);
      Ixr[0][j]=-LARGE;
      Iyr[0][j]=((!tgw) && (e2==ali2->len) ? 0.0f : -go-(float)(m2-j)*ge);
    }
  
  /* iterate over first alignment */
  for (i=l1-1; i>=0; i--)
    {
      /* iterate over second alignment */
      for (j=l2-1; j>=0; j--)
	{
	  /* set pre-factors */
	  f_i=((!tgw) && (s1==0) && (i==0) ? 0.0f : 1.0f);
	  f_j=((!tgw) && (s2==0) && (j==0) ? 0.0f : 1.0f);
	  
	  DPr[i][j]=DPr[i+1][j+1]+S[s1+i][s2+j]; /* add pairwise score */
	  Ixr[i][j]=MAX (DPr[i+1][j]-f_j*go, Ixr[i+1][j]-f_j*ge);
	  Iyr[i][j]=MAX (DPr[i][j+1]-f_i*go, Iyr[i][j+1]-f_i*ge);
	  maxarg=MAXARG3 (DPr[i][j], Ixr[i][j], Iyr[i][j]);
	  if (maxarg==0) /* DPr[i][j]=DPr[i][j] */;
	  else if (maxarg==1) DPr[i][j]=Ixr[i][j];
	  else DPr[i][j]=Iyr[i][j];
	}
    }
  
  /* calculate optimal alignment score and column of additional cost matrix */
  score=-LARGE;
  for (i=0; i<m1; i++)
    {
      for (j=0; j<m2; j++)
	{
	  /* set pre-factors */
	  f_i=((!tgw) && (((i==0) && (s1==0)) || ((i==l1) && (e1==ali1->len))) ? 0.0f : 1.0f);
	  f_j=((!tgw) && (((j==0) && (s2==0)) || ((j==l2) && (e2==ali2->len))) ? 0.0f : 1.0f);
	  
	  C[i][j]=MAX3 (DPf[i][j]+DPr[i][j], Ixf[i][j]+Ixr[i][j]+f_j*(go-ge), Iyf[i][j]+Iyr[i][j]+f_i*(go-ge));
	  if (C[i][j]>score) score=C[i][j]; /* get maximal matrix element */
	}
    }
  
  /* normalize scores */
  for (i=0; i<m1; i++) for (j=0; j<m2; j++) C[i][j]-=score; /* substract maximum score -> so matrix elemtents are less or equal zero */

  /* free memory */
  FREE2D (DPf, m1, i);
  FREE2D (Ixf, m1, i);
  FREE2D (Iyf, m1, i);
  FREE2D (DPr, m1, i);
  FREE2D (Ixr, m1, i);
  FREE2D (Iyr, m1, i);

  return (C);
}

/*******************************************************
calculates optimal slicing positions
*******************************************************/
void calc3slicing (float **Cxy, float **Cxz, float **Cyz, int l1, int l2, int l3, int *p1, int *p2, int *p3)
{
  int c1, c2, c3;
  int i, j, k;
  float Ccur;
  float Cmax=-FLARGE;
  int *idx;

  /* get data */
  int m1=l1+1;
  int m2=l2+1;
  int m3=l3+1;
  int maxlen=MAX3 (m1, m2, m3);

  /* create a sequence 0,+1,-1,+2,-2,+3,... */
  MAKE1DINT (idx, maxlen);
  for (i=0; i<maxlen; i++)
    {
      if (i%2) idx[i]=(i+1)/2;
      else idx[i]=-i/2;
    }

  if (maxlen==m1) /* first sequence is longest */
    {
      c1=l1/2;
      /* search for optimal slice positions */
      for (j=0; j<m2; j++)
	{
	  c2=l2/2+idx[j];
	  Ccur=Cxy[c1][c2];
	  /* addition score so far above maximum additional score */
	  if (Ccur>Cmax) for (k=0; k<m3; k++)
	    {
	      c3=l3/2+idx[k];
	      Ccur=Cxy[c1][c2]+Cxz[c1][c3]+Cyz[c2][c3];
	      /* new maximum found */
	      if (Ccur>Cmax)
		{
		  Cmax=Ccur;
		  *p1=c1;
		  *p2=c2;
		  *p3=c3;
		}
	    }
	}
    }
  else if (maxlen==m2) /* second sequence is longest */
    {
      c2=l2/2;
      /* search for optimal slice positions */
      for (i=0; i<m1; i++)
	{
	  c1=l1/2+idx[i];
	  Ccur=Cxy[c1][c2];
	  if (Ccur>Cmax) for (k=0; k<m3; k++)
	    {
	      c3=l3/2+idx[k];
	      Ccur=Cxy[c1][c2]+Cxz[c1][c3]+Cyz[c2][c3];
	      /* new maximum found */
	      if (Ccur>Cmax)
		{
		  Cmax=Ccur;
		  *p1=c1;
		  *p2=c2;
		  *p3=c3;
		}
	    }
	}
    }
  else /* third sequence is longest */
    {
      c3=l3/2;
      /* search for optimal slice positions */
      for (i=0; i<m1; i++)
	{
	  c1=l1/2+idx[i];
	  Ccur=Cxz[c1][c3];
	  if (Ccur>Cmax) for (j=0; j<m2; j++)
	    {
	      c2=l2/2+idx[j];
	      Ccur=Cxy[c1][c2]+Cxz[c1][c3]+Cyz[c2][c3];
	      /* new maximum found */
	      if (Ccur>Cmax)
		{
		  Cmax=Ccur;
		  *p1=c1;
		  *p2=c2;
		  *p3=c3;
		}
	    }
	}
    }

  /* free memory */
  free (idx);

#if DEBUG
  fprintf (debugfile, "calc3slicing: slicing positions (%d, %d, %d)\n", *p1, *p2, *p3);
#endif 
}
