#include <stdlib.h>
#include <stdio.h>

#include "list.h"

/*******************************************************
implementation of an doubly linked list with dummy 
elements
*******************************************************/
struct S_LIST {
  int num;
  LISTNODE* begin;
  LISTNODE* end;
};

struct S_LISTNODE {
  Item item;
  LISTNODE* prev;
  LISTNODE* next;
};


/*******************************************************
make a new list node
********************************************************/
LISTNODE* LISTNODEinit (Item item, LISTNODE* prev, LISTNODE* next)
{
  LISTNODE* n=(LISTNODE*)malloc ((size_t)sizeof (LISTNODE));
  n->item=item;
  n->prev=prev;
  n->next=next;
  return (n);
}

/*******************************************************
initializes the list
*******************************************************/
LIST* LISTinit (void)
{
  LIST* l=(LIST*)malloc ((size_t)sizeof (LIST));
  l->num=0;
  l->begin=LISTNODEinit (0, NULL, NULL);
  l->end=LISTNODEinit (0, NULL, NULL);
  l->begin->next=l->end;
  l->end->prev=l->begin;
  return (l);
}

/*******************************************************
add a node to the end of a list
*******************************************************/
void LISTpush (LIST* l, Item item)
{
  LISTNODE* n=LISTNODEinit (item, l->end->prev, l->end);
  l->end->prev->next=n;
  l->end->prev=n;
  ++l->num;	
}

/*******************************************************
removes the last node of a list and return it
*******************************************************/
Item LISTpop (LIST* l)
{
  LISTNODE* n;
  Item item;
  if (l->num)
    {
      n=l->end->prev;
      item=n->item;
      n->prev->next=n->next;
      n->next->prev=n->prev;
      free (n);
      --l->num;
      return (item);
    }
  return (0);
}

/*******************************************************
add a node to the end of a list
*******************************************************/
void LISTput (LIST* l, Item item)
{
  LISTNODE* n=LISTNODEinit (item, l->end->prev, l->end);
  l->end->prev->next=n;
  l->end->prev=n;
  ++l->num;
}

/*******************************************************
removes the first node of a list and returns it
*******************************************************/
Item LISTget (LIST* l)
{
  LISTNODE* n;
  Item item;
  if (l->num)
    {
      n=l->begin->next;
      item=n->item;
      l->begin->next=n->next;
      n->next->prev=n->prev;
      free (n);
      --l->num;
      return (item);
    }
  return (0);
}

/*******************************************************
add a node at a given position to the list
*******************************************************/
void LISTinsert (LIST* l, Item item, int pos)
{
  LISTNODE* n;
  if (pos<=l->num)
    {
      if (pos<l->num/2)
	{
	  n=l->begin;
	  while (pos--) n=n->next;
	  n->next=LISTNODEinit (item, n, n->next);
	  n->next->next->prev=n->next;
	}
      else
	{
	  n=l->end;
	  while (l->num-pos++) n=n->prev;
	  n->prev=LISTNODEinit (item, n->prev, n);
	  n->prev->prev->next=n->prev;
	}
      ++l->num;	
    }
}

/*******************************************************
removes the element at a given position from the list
and return it
********************************************************/
Item LISTremove (LIST* l, int pos)
{
  LISTNODE* n;
  Item item;
  if (pos<=l->num)
    {
      if (pos<l->num/2)
	{
	  n=l->begin->next;
	  while (pos--) n=n->next;
	}
      else
	{
	  n=l->end->prev;
	  while (l->num-pos++-1) n=n->prev;
	}
      n->prev->next=n->next;
      n->next->prev=n->prev;
      item=n->item;
      free (n);
      --l->num;
      return (item);
    }
  return (0);
}

/*******************************************************
return the first element
*******************************************************/
Item LISTfirst (LIST* l)
{
  return (l->begin->next);
}

Item LISTfront (LIST* l)
{
  return (l->begin->next);
}

/*******************************************************
return the last element
*******************************************************/
Item LISTlast (LIST* l)
{
  return (l->end->prev);
}

Item LISTback (LIST* l)
{
  return (l->end->prev);
}

/*******************************************************
return the item at a given position
*******************************************************/
Item LISTitem (LIST* l, int pos)
{
  LISTNODE* n;
  if (pos<l->num/2)
    {
      n=l->begin->next;
      while (pos--) n=n->next;
    }
  else
    {
      n=l->end->prev;
      while (l->num-pos++) n=n->prev;
    }
  return (n->item);
}

/*******************************************************
check whether list is empty
*******************************************************/
int LISTempty (LIST* l)
{
  return (l->num==0);
}

/*******************************************************
returns the number of items in the list
*******************************************************/
int LISTnum (LIST* l)
{
  return (l->num);
}

/*******************************************************
remove all items from the list
*******************************************************/
void LISTclear (LIST* l)
{
  LISTNODE* tmp=l->begin->next;
  LISTNODE* next;
  while (tmp->next!=NULL)
    {
      next=tmp->next;
      free (tmp);
      tmp=next;
    }
  l->begin->next=l->end;
  l->end->prev=l->begin;
  l->num=0;
}

/*******************************************************
frees the list
*******************************************************/
void LISTfree (LIST* l)
{
  LISTNODE* tmp=l->begin->next;
  LISTNODE* next;
  while (tmp->next!=NULL)
    {
      next=tmp->next;
      free (tmp);
      tmp=next;
    }
  l->begin->next=l->end;
  l->end->prev=l->begin;
  free (l);
}
