/**********************************************************
defines a set of macros for time critical housekeeping tasks
***********************************************************/
#ifndef MACROS_H
#define MACROS_H

/* allocate memory of given type and number of */
/* elements */
#define MALLOC(P,T,S) { if((P=(T*)malloc(sizeof(T)*(size_t)(S)))==NULL) { \
      fprintf(stderr,"malloc failed.\n"); \
      exit(EXIT_FAILURE); \
    } memset(P,0,sizeof(T)*(size_t)(S)); \
  }

/* re-allocate memory of given type and number of */
/* elements */
#define REALLOC(P,T,S) { if((P=(T*)realloc(P,sizeof(T)*(size_t)(S)))==NULL) { \
      fprintf(stderr,"realloc failed.\n");				\
      exit(EXIT_FAILURE); \
    } }

/* allocates an array of chars */
#define MAKE1DCHAR(D,n1) { MALLOC(D,char,n1); }

/* allocates a two-dimensional array of chars */
#define MAKE2DCHAR(D,n1,n2,i) { MALLOC(D,char*,n1); \
    for(i=0;i<n1;i++) {	\
      MALLOC(D[i],char,n2); \
    } }

/* allocates a three-dimensional array of chars */
#define MAKE3DCHAR(D,n1,n2,n3,i,j) { MALLOC(D,char**,n1); \
    for(i=0;i<n1;i++) {	\
      MALLOC(D[i],char*,n2); \
      for (j=0;j<n2;j++) { \
	MALLOC(D[i][j],char,n3); \
      } } }

/* allocates an array of ints */
#define MAKE1DINT(D,n1)	{ MALLOC(D,int,n1); }

/* allocates an two-dimensional array of ints */
#define MAKE2DINT(D,n1,n2,i) { MALLOC(D,int*,n1); \
    for(i=0;i<n1;i++) {	\
      MALLOC(D[i],int,n2);			\
    } }

/* allocates an array of floats */
#define MAKE1DFLOAT(D,n1) { MALLOC(D,float,n1); }

/* allocates an two-dimensional array of floats */
#define MAKE2DFLOAT(D,n1,n2,i) { MALLOC(D,float*,n1); \
    for(i=0;i<n1;i++) {	\
      MALLOC(D[i],float,n2); \
    } }

/* allocates a three-dimensional array of floats */
#define MAKE3DFLOAT(D,n1,n2,n3,i,j) { MALLOC(D,float**,n1); \
    for(i=0;i<n1;i++) {	\
      MALLOC(D[i],float*,n2); \
      for (j=0;j<n2;j++) { \
	MALLOC(D[i][j],float,n3); \
      } } }

/* allocates an array of doubles */
#define MAKE1DDOUBLE(D,n1) { MALLOC(D,double,n1); }

/* allocates an two-dimensional array of doubles */
#define MAKE2DDOUBLE(D,n1,n2,i)	{ MALLOC(D,double*,n1); \
    for(i=0;i<n1;i++) {	\
      MALLOC(D[i],double,n2); \
    } }

/* frees an array of arbitrary type */
#define FREE(D)	{ free(D); D=NULL; }

/* frees an array of arbitrary type */
#define FREE1D(D) { free(D); D=NULL; }

/* frees an two-diemensional array of arbitrary type */
#define FREE2D(D,n1,i) { for(i=0;i<n1;i++) free(D[i]); \
    free (D); D=NULL; }

/* frees an three-diemensional array of arbitrary type */
#define FREE3D(D,n1,n2,i,j) { for(i=0;i<n1;i++) { \
      for (j=0;j<n2;++j) free(D[i][j]);	\
      free(D[i]); }			\
    free (D); D=NULL; }

/* returns the minimum of two values */
#define MIN(a,b) (((a)<=(b))?(a):(b))

/* returns the maximum of two values */
#define MAX(a,b) (((a)>=(b))?(a):(b))

/* returns the minimum of three values */
#define MIN3(a,b,c) (((a)<=(b))?(((a)<=(c))?(a):(c)):((b)<=(c))?(b):(c))

/* returns the maximum of three values */
#define MAX3(a,b,c) (((a)>=(b))?(((a)>=(c))?(a):(c)):((b)>=(c))?(b):(c))

/* returns the number of the minimum argument of two values */
#define MINARG(a,b) (((a)<=(b))?0:1)

/* returns the number of the maximum argument of two values */
#define MAXARG(a,b) (((a)>=(b))?0:1)

/* returns the number of the minimum of three values */
#define MINARG3(a,b,c) (((a)<=(b))?(((a)<=(c))?0:2):((b)<=(c))?1:2)

/* returns the number of the maximum of three values */
#define MAXARG3(a,b,c) (((a)>=(b))?(((a)>=(c))?0:2):((b)>=(c))?1:2)

/* prints an error message to stderr and terminate execution */
#define FATAL(msg) { fprintf(stderr,"Error: %s\n\n",msg); exit (EXIT_FAILURE); }

/* prints a warning message to stderr */
#define WARN(msg) { fprintf(stderr,"Warning: %s\n\n",msg); }

/* swaps two items */
#define SWAP(a,b,tmp) { tmp=a; a=b; b=tmp; }

/* swap two items in an array */
#define SWAP_TUPLE(t,e1,e2,tmp) { tmp=t[e1]; t[e1]=t[e2]; t[e2]=tmp; }

/* returns random number in intervall [0,A) */
#define RANDINT(A) ((int)((A)*rand()/(RAND_MAX+1.0)))

#endif
