#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "global.h"
#include "options.h"
#include "seqio.h"
#include "macros.h"
#include "scoring.h"

extern OPTION Options;
extern SEQ Seqs;

/********************************************************
make a new scoring set
********************************************************/
SCORING *makeScoring (void)
{
  int i;
  SCORING *scoring;

  MALLOC (scoring, SCORING, 1);
  MAKE2DFLOAT (scoring->M, NUMLET, NUMLET, i);

  return (scoring);
}

/********************************************************
determine sequence type and set correct gap penalties
convert to lower case
*********************************************************/
void setupScoring (SCORING *D_scoring, SCORING *A_scoring)
{
  /* set distance scoring */
  if (D_scoring)
    {
      D_scoring->seqtype=Seqs.type;
      D_scoring->gaptype=GAP_2D;
      D_scoring->termgapweight=FALSE; /* for distance calculation weight end gaps full */
      D_scoring->usestruc=FALSE;
      D_scoring->psi=0.0f;
  
      /* set correct gap penalties */
      if ((Seqs.type==SEQ_DNA) || (Seqs.type==SEQ_RNA))
	{
	  D_scoring->go=(Options.D_Gogiven ? Options.D_Gapopen : D_NUCGAPOPEN);
	  D_scoring->ge=(Options.D_Gegiven ? Options.D_Gapextend : D_NUCGAPEXTEND);
	}
      else if (Seqs.type==SEQ_PROTEIN)
	{
	  D_scoring->go=(Options.D_Gogiven ? Options.D_Gapopen : A_AAGAPOPEN);
	  D_scoring->ge=(Options.D_Gegiven ? Options.D_Gapextend : A_AAGAPEXTEND);
	}
      else FATAL ("Unknow sequence type.");
    }
  
  /* set alignment scoring */
  if (A_scoring)
    {
      A_scoring->seqtype=Seqs.type;
      A_scoring->gaptype=Options.Gaptype;
      A_scoring->termgapweight=Options.Termgapweight;
      A_scoring->usestruc=Options.Usestruc;
      A_scoring->psi=Options.Psi;
      
      /* set correct gap penalties */
      if ((Seqs.type==SEQ_DNA) || (Seqs.type==SEQ_RNA))
	{
	  A_scoring->go=(Options.A_Gogiven ? Options.A_Gapopen : D_NUCGAPOPEN);
	  A_scoring->ge=(Options.A_Gegiven ? Options.A_Gapextend : D_NUCGAPEXTEND);
	}
      else if (Seqs.type==SEQ_PROTEIN)
	{
	  A_scoring->go=(Options.A_Gogiven ? Options.A_Gapopen : A_AAGAPOPEN);
	  A_scoring->ge=(Options.A_Gegiven ? Options.A_Gapextend : A_AAGAPEXTEND);
	}
      else FATAL ("Unknow sequence type.");
    
      /* is secondary structure annotation possible? */
      if ((Options.Usestruc) && (Seqs.type!=SEQ_RNA))
	{
	  WARN ("Disabling secondary structure annotation for DNA or protein sequences.");
	  A_scoring->usestruc=FALSE;
	}
    }
}
