#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "global.h"
#include "split.h"
#include "macros.h"

#ifdef DEBUG
extern FILE *debugfile;
#endif

/*******************************************************
splits the alignment of three profiles into two
alignments of two profiles
*******************************************************/
void splitAli (ALI *ali1, ALI *ali2, ALI *ali, int u, int v, int n1, int n2, int n3, int *did1, int *did2, int *ndid)
{
  /* dids1: indices of duplicated sequences in first alignment */
  /* dids2: indices of duplicated sequences in second alignment */
  /* ndid: number of duplicated sequences */
  int i;
  int deslen;

  /* get data */
  int alilen=ali->len;

  /* split the alignment XYZ into two alignments XY and YZ */
  /* make alignment XY */
  ali1->id=u;
  ali1->num=n1+n2;
  ali1->len=alilen;
  ali1->score=0.0f;
  MALLOC (ali1->descriptors, char*, n1+n2);
  MALLOC (ali1->sequences, char*, n1+n2);
  MALLOC (ali1->sinfo, SINFO*, n1+n2);
  MALLOC (ali1->sids, int, n1+n2);
  for (i=0; i<(n1+n2); i++)
    {
      deslen=(int)strlen (ali->descriptors[i]);
      MALLOC (ali1->descriptors[i], char, deslen+1);
      MALLOC (ali1->sequences[i], char, alilen+1);
      MALLOC (ali1->sinfo[i], SINFO, alilen);
      strncpy (ali1->descriptors[i], ali->descriptors[i], deslen+1);
      strncpy (ali1->sequences[i], ali->sequences[i], alilen+1);
      memcpy  (ali1->sinfo[i], ali->sinfo[i], (size_t)alilen*sizeof (SINFO)); 
      ali1->sids[i]=ali->sids[i];
    }

  /* make alignment YZ */
  ali2->id=v;
  ali2->num=n2+n3;
  ali2->len=alilen;
  ali2->score=0.0f;
  MALLOC (ali2->descriptors, char*, n2+n3);
  MALLOC (ali2->sequences, char*, n2+n3);
  MALLOC (ali2->sinfo, SINFO*, n2+n3);
  MALLOC (ali2->sids, int, n2+n3);
  for (i=0; i<(n2+n3); i++)
    {
      deslen=(int)strlen (ali->descriptors[i+n1]);
      MALLOC (ali2->descriptors[i], char, deslen+1);
      MALLOC (ali2->sequences[i], char, alilen+1);
      MALLOC (ali2->sinfo[i], SINFO, alilen);
      strncpy (ali2->descriptors[i], ali->descriptors[i+n1], deslen+1);
      strncpy (ali2->sequences[i], ali->sequences[i+n1], alilen+1);
      memcpy (ali2->sinfo[i], ali->sinfo[i+n1], (size_t)alilen*sizeof (SINFO));
      ali2->sids[i]=ali->sids[i+n1];
    }

  /* fill array with ids of double occuring sequences */
  *ndid=n2; /* sequences in Y are doubled */
  for (i=0; i<n2; i++)
    {
      did1[i]=n1+i;
      did2[i]=i;
    }

#ifdef DEBUG
  fprintf (debugfile, "split: alignments after splitting:\n");
  fprintf (debugfile, "ali1: #seqs=%d, len=%d\n", ali1->num, ali1->len);
  for (i=0; i<ali1->num; i++) fprintf (debugfile, "%s: %s\n", ali1->descriptors[i], ali1->sequences[i]);
  fprintf (debugfile, "ali2: #seqs=%d, len=%d\n", ali2->num, ali2->len);
  for (i=0; i<ali2->num; i++) fprintf (debugfile, "%s: %s\n", ali2->descriptors[i], ali2->sequences[i]);
#endif
}
