#include <stdlib.h>
#include <string.h>
#include <stdio.h>

/* Vienna RNA */
#include "ViennaRNA/fold_vars.h"
#include "ViennaRNA/part_func.h"

#include "global.h"
#include "options.h"
#include "sinfo.h"
#include "sprob.h"
#include "macros.h"

#ifdef DEBUG
extern FILE *debugfile;
#endif

/***********************************************************
get the structure information for a given sequence 
************************************************************/
SINFO *getStruct (char *sequence, BOOL usestruc)
{
  int i, j;
  SINFO *sinfo=NULL;

  /* get data */
  int len=(int)strlen (sequence);
  
  /* allocate memory */
  MALLOC (sinfo, SINFO, len);

  if (usestruc)
    {
      /* allocates memory for folding sequence */
      init_pf_fold (len);

      /* enable backtracking to compute pairwise probabilities */
      do_backtrack=1;
      
      /* fold sequence and calculate pair probability */
      pf_fold (sequence, NULL);

#ifdef DEBUG
      fprintf (debugfile, "pair probabilities\n");
      for (j=1; j<=len; j++) for (i=1; i<j; i++) fprintf (debugfile, "%3d %3d: %8.6f\n", i, j, pr[iindx[i]-j]);
#endif

      for (j=1; j<=len; j++) 
	{
	  sinfo[j-1].unpaired=1.0f;
	  sinfo[j-1].dnpaired=0.0f;
	  sinfo[j-1].uppaired=0.0f;
	  for (i=1; i<j; i++)
	    {
	      sinfo[j-1].dnpaired+=pr[iindx[i]-j]; /* j is paired downstream */
	      sinfo[j-1].unpaired-=pr[iindx[i]-j]; /* j is unpaired */
	    }
	  for (i=j+1; i<=len; i++) 
	    {
	      sinfo[j-1].uppaired+=pr[iindx[j]-i]; /* j is paired upstream */
	      sinfo[j-1].unpaired-=pr[iindx[j]-i]; /* j is unpaired */
	    }
	}
      
      /* free allocated memory */
      free_pf_arrays ();
    }

#ifdef DEBUG
  if (usestruc)
    {
      fprintf (debugfile, "getStruct\n");
      for (i=0; i<len; i++) fprintf (debugfile, "%3d: down=%8.6f up=%8.6f un=%8.6f\n", i+1, sinfo[i].dnpaired, sinfo[i].uppaired, sinfo[i].unpaired);
    }
#endif

  return (sinfo);
}
